from __future__ import annotations

import logging
import os
import pathlib
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Iterable, Union

from ada.base.changes import ChangeAction
from ada.base.ifc_types import SpatialTypes
from ada.base.physical_objects import BackendGeom
from ada.base.types import GeomRepr
from ada.base.units import Units
from ada.cache.store import CacheStore
from ada.concepts.connections import JointBase
from ada.concepts.containers import (
    Beams,
    Connections,
    Materials,
    Nodes,
    Plates,
    Sections,
)
from ada.concepts.groups import Group
from ada.concepts.piping import Pipe
from ada.concepts.points import Node
from ada.concepts.presentation_layers import PresentationLayers
from ada.concepts.primitives import (
    Penetration,
    PrimBox,
    PrimCyl,
    PrimExtrude,
    PrimRevolve,
    Shape,
)
from ada.concepts.transforms import Instance, Placement
from ada.concepts.user import User
from ada.config import Settings, get_logger
from ada.fem import (
    Connector,
    Csys,
    FemSet,
    Mass,
    StepEigen,
    StepExplicit,
    StepImplicit,
    StepSteadyState,
)
from ada.fem.concept import FEM

if TYPE_CHECKING:
    import ifcopenshell

    from ada import Beam, Material, Plate, Section, Wall, Weld
    from ada.fem.containers import COG
    from ada.fem.formats.general import FEATypes, FemConverters
    from ada.fem.meshing import GmshOptions
    from ada.fem.results.common import FEAResult
    from ada.ifc.store import IfcStore
    from ada.visualize.concept import VisMesh
    from ada.visualize.config import ExportConfig

_step_types = Union[StepSteadyState, StepEigen, StepImplicit, StepExplicit]

logger = get_logger()


class FormatNotSupportedException(Exception):
    pass


@dataclass
class _ConvertOptions:
    ecc_to_mpc: bool = True
    hinges_to_coupling: bool = True

    # From FEM to concepts
    fem2concepts_include_ecc = False


class Part(BackendGeom):
    """A Part superclass design to host all relevant information for cad and FEM modelling."""

    IFC_CLASSES = SpatialTypes

    def __init__(
        self,
        name,
        colour=None,
        placement=Placement(),
        fem: FEM = None,
        settings: Settings = Settings(),
        metadata=None,
        parent=None,
        units: Units = Units.M,
        guid=None,
        ifc_store: IfcStore = None,
        ifc_class: SpatialTypes = SpatialTypes.IfcBuildingStorey,
    ):
        super().__init__(name, guid=guid, metadata=metadata, units=units, parent=parent, ifc_store=ifc_store)
        self._nodes = Nodes(parent=self)
        self._beams = Beams(parent=self)
        self._plates = Plates(parent=self)
        self._pipes = list()
        self._walls = list()
        self._connections = Connections(parent=self)
        self._materials = Materials(parent=self)
        self._sections = Sections(parent=self)
        self._colour = colour
        self._placement = placement
        self._instances: dict[Any, Instance] = dict()
        self._shapes = []
        self._welds = []
        self._parts = dict()
        self._groups: dict[str, Group] = dict()
        self._ifc_class = ifc_class
        self._props = settings
        if fem is not None:
            fem.parent = self

        self.fem = FEM(name + "-1", parent=self) if fem is None else fem

    def add_beam(self, beam: Beam, add_to_layer: str = None) -> Beam:
        if beam.units != self.units:
            beam.units = self.units
        beam.parent = self

        mat = self.add_material(beam.material)
        if mat != beam.material:
            beam.material = mat

        sec = self.add_section(beam.section)
        if sec != beam.section:
            beam.section = sec

        tap = self.add_section(beam.taper)
        if tap != beam.taper:
            beam.taper = tap

        old_node = self.nodes.add(beam.n1)
        if old_node != beam.n1:
            beam.n1 = old_node

        old_node = self.nodes.add(beam.n2)
        if old_node != beam.n2:
            beam.n2 = old_node

        beam.change_type = beam.change_type.ADDED
        self.beams.add(beam)

        if add_to_layer is not None:
            a = self.get_assembly()
            a.presentation_layers.add_object(beam, add_to_layer)

        return beam

    def add_plate(self, plate: Plate, add_to_layer: str = None) -> Plate:
        if plate.units != self.units:
            plate.units = self.units

        plate.parent = self

        mat = self.add_material(plate.material)
        if mat is not None:
            plate.material = mat

        for n in plate.nodes:
            self.nodes.add(n)

        plate.change_type = plate.change_type.ADDED
        self._plates.add(plate)

        if add_to_layer is not None:
            a = self.get_assembly()
            a.presentation_layers.add_object(plate, add_to_layer)

        return plate

    def add_pipe(self, pipe: Pipe, add_to_layer: str = None) -> Pipe:
        if pipe.units != self.units:
            pipe.units = self.units
        pipe.parent = self

        mat = self.add_material(pipe.material)
        if mat != pipe.material:
            pipe.material = mat
            mat.refs.append(pipe)

        sec = self.add_section(pipe.section)
        if sec != pipe.section:
            pipe.section = sec
            sec.refs.append(pipe)

        for seg in pipe.segments:
            mat = self.add_material(seg.material)
            if mat != seg.material:
                seg.material = mat
                mat.refs.append(seg)

            sec = self.add_section(seg.section)
            if sec != seg.section:
                seg.section = sec
                sec.refs.append(seg)

        pipe.change_type = pipe.change_type.ADDED
        self.pipes.append(pipe)

        if add_to_layer is not None:
            a = self.get_assembly()
            a.presentation_layers.add_object(pipe, add_to_layer)

        return pipe

    def add_wall(self, wall: Wall) -> Wall:
        if wall.units != self.units:
            wall.units = self.units
        wall.parent = self
        self._walls.append(wall)
        return wall

    def add_shape(self, shape: Shape, change_type: ChangeAction = ChangeAction.ADDED) -> Shape:
        if shape.units != self.units:
            logger.info(f'shape "{shape}" has different units. changing from "{shape.units}" to "{self.units}"')
            shape.units = self.units
        shape.parent = self

        mat = self.add_material(shape.material)
        if mat != shape.material:
            shape.material = mat

        shape.change_type = change_type
        self._shapes.append(shape)
        return shape

    def add_part(self, part: Part, overwrite: bool = False, add_to_layer: str = None) -> Part:
        if issubclass(type(part), Part) is False:
            raise ValueError("Added Part must be a subclass or instance of Part")

        if part.units != self.units:
            part.units = self.units
        part.parent = self

        if part.name in self._parts.keys() and overwrite is False:
            raise ValueError(f'Part name "{part.name}" already exists. Pass "overwrite=True" to replace existing part.')

        self._parts[part.name] = part
        try:
            part._on_import()
        except NotImplementedError:
            logger.info(f'Part "{part}" has not defined its "on_import()" method')

        part.change_type = part.change_type.ADDED
        if add_to_layer is not None:
            a = self.get_assembly()
            a.presentation_layers.add_object(part, add_to_layer)

        return part

    def add_joint(self, joint: JointBase) -> JointBase:
        """
        This method takes a Joint element containing two intersecting beams. It will check with the existing
        list of joints to see whether or not it is part of a larger more complex joint. It usese primarily
        two criteria.

        Criteria 1: If both elements are in an existing joint already, it will u

        Criteria 2: If the intersecting point coincides within a specified tolerance (currently 10mm)
        with an exisiting joint intersecting point. If so it will add the elements to this joint.
        If not it will create a new joint based on these two members.
        """
        if joint.units != self.units:
            joint.units = self.units
        self._connections.add(joint)
        return joint

    def add_weld(self, weld: Weld) -> Weld:
        weld.parent = self
        self._welds.append(weld)

        return weld

    def add_material(self, material: Material) -> Material:
        if material.units != self.units:
            material.units = self.units
        material.parent = self
        return self._materials.add(material)

    def add_section(self, section: Section) -> Section:
        if section.units != self.units:
            section.units = self.units
        return self._sections.add(section)

    def add_object(self, obj: Part | Beam | Plate | Wall | Pipe | Shape | Weld):
        from ada import Beam, Part, Pipe, Plate, Shape, Wall, Weld

        if isinstance(obj, Beam):
            self.add_beam(obj)
        elif isinstance(obj, Plate):
            self.add_plate(obj)
        elif isinstance(obj, Pipe):
            self.add_pipe(obj)
        elif issubclass(type(obj), Part):
            self.add_part(obj)
        elif issubclass(type(obj), Shape):
            self.add_shape(obj)
        elif isinstance(obj, Wall):
            self.add_wall(obj)
        elif isinstance(obj, Weld):
            self.add_weld(obj)
        else:
            raise NotImplementedError(f'"{type(obj)}" is not yet supported for smart append')

    def add_penetration(
        self,
        pen: Penetration | PrimExtrude | PrimRevolve | PrimCyl | PrimBox,
        add_pen_to_subparts=True,
        add_to_layer: str = None,
    ) -> Penetration:
        def create_pen(pen_):
            if isinstance(pen_, (PrimExtrude, PrimRevolve, PrimCyl, PrimBox)):
                return Penetration(pen_, parent=self)
            return pen_

        for bm in self.beams:
            bm.add_penetration(create_pen(pen), add_to_layer=add_to_layer)

        for pl in self.plates:
            pl.add_penetration(create_pen(pen), add_to_layer=add_to_layer)

        for shp in self.shapes:
            shp.add_penetration(create_pen(pen), add_to_layer=add_to_layer)

        for pipe in self.pipes:
            for seg in pipe.segments:
                seg.add_penetration(create_pen(pen), add_to_layer=add_to_layer)

        for wall in self.walls:
            wall.add_penetration(create_pen(pen), add_to_layer=add_to_layer)

        if add_pen_to_subparts:
            for p in self.get_all_subparts():
                p.add_penetration(pen, False, add_to_layer=add_to_layer)

        return pen

    def add_instance(self, element, placement: Placement):
        if element not in self._instances.keys():
            self._instances[element] = Instance(element)
        self._instances[element].placements.append(placement)

    def add_set(self, name, set_members: list[Part | Beam | Plate | Wall | Pipe | Shape]) -> Group:
        if name not in self.groups.keys():
            self.groups[name] = Group(name, set_members, parent=self)
        else:
            logger.info(f'Appending set "{name}"')
            for mem in set_members:
                if mem not in self.groups[name].members:
                    self.groups[name].members.append(mem)

        return self.groups[name]

    def add_elements_from_ifc(self, ifc_file_path: os.PathLike | str, data_only=False):
        from ada import Beam, Pipe, Plate, Shape, Wall

        a = Assembly("temp")
        a.read_ifc(ifc_file_path, data_only=data_only)
        for sec in a.get_all_sections():
            res = self.add_section(sec)
            if res == sec:
                sec.change_type = ChangeAction.ADDED

        for mat in a.get_all_materials():
            res = self.add_material(mat)
            if res == mat:
                mat.change_type = ChangeAction.ADDED

        for obj in a.get_all_physical_objects():
            if issubclass(type(obj), Shape):
                self.add_shape(obj)
            elif isinstance(obj, Beam):
                self.add_beam(obj)
            elif isinstance(obj, Plate):
                self.add_plate(obj)
            elif isinstance(obj, Wall):
                self.add_wall(obj)
            elif isinstance(obj, Pipe):
                self.add_pipe(obj)
            else:
                raise ValueError(f"Unrecognized {type(obj)=}")

    def read_step_file(
        self,
        step_path,
        name=None,
        scale=None,
        transform=None,
        rotate=None,
        colour=None,
        opacity=1.0,
        source_units=Units.M,
        include_shells=False,
    ):
        """

        :param step_path: Can be path to stp file or path to directory of step files.
        :param name: Desired name of destination Shape object
        :param scale: Scale the step content upon import
        :param transform: Transform the step content upon import
        :param rotate: Rotate step content upon import
        :param colour: Assign a specific colour upon import
        :param opacity: Assign Opacity upon import
        :param source_units: Unit of the imported STEP file. Default is 'm'
        """
        from ada.occ.utils import extract_shapes

        shapes = extract_shapes(step_path, scale, transform, rotate, include_shells=include_shells)

        if len(shapes) > 0:
            ada_name = name if name is not None else "CAD" + str(len(self.shapes) + 1)
            for i, shp in enumerate(shapes):
                ada_shape = Shape(ada_name + "_" + str(i), shp, colour, opacity, units=source_units)
                self.add_shape(ada_shape)

    def calculate_cog(self) -> COG:
        import numpy as np

        from ada import Beam, Plate, Shape
        from ada.core.vector_utils import (
            local_2_global_points,
            poly2d_center_of_gravity,
            poly_area_from_list,
        )
        from ada.fem.containers import COG

        tot_mass = 0
        cogs = []
        for obj in self.get_all_physical_objects():
            if issubclass(type(obj), Shape):  # Assuming Mass & COG is manually assigned to arbitrary shape
                cogs.append(np.array(obj.cog) * obj.mass)
                tot_mass += obj.mass
            elif isinstance(obj, Beam):
                rho = obj.material.model.rho
                area = obj.section.properties.Ax
                length = obj.length
                mass = rho * area * length
                cog = (obj.n1.p + obj.n2.p) / 2
                cogs.append(cog * mass)
                tot_mass += mass
            elif isinstance(obj, Plate):
                rho = obj.material.model.rho
                positions = np.array(obj.poly.points2d)
                place = obj.poly.placement

                area = poly_area_from_list(obj.poly.points2d)
                cog2d = poly2d_center_of_gravity(positions)
                cog = local_2_global_points([cog2d], place.origin, place.xdir, place.zdir)[0]
                mass = rho * obj.t * area
                cogs.append(cog * mass)
                tot_mass += mass

        for mass in self.fem.masses.values():
            cogs.append(np.array(mass.nodes[0].p) * mass.mass)
            tot_mass += mass.mass

        cog = sum(cogs) / tot_mass
        return COG(cog, tot_mass)

    def create_objects_from_fem(self, skip_plates=False, skip_beams=False) -> None:
        """Build Beams and Plates from the contents of the local FEM object"""
        from ada.fem.formats.utils import convert_part_objects

        if type(self) is Assembly:
            for p_ in self.get_all_parts_in_assembly():
                logger.info(f'Beginning conversion from fem to structural objects for "{p_.name}"')
                convert_part_objects(p_, skip_plates, skip_beams)
        else:
            logger.info(f'Beginning conversion from fem to structural objects for "{self.name}"')
            convert_part_objects(self, skip_plates, skip_beams)
        logger.info("Conversion complete")

    def get_part(self, name: str) -> Part:
        key_map = {key.lower(): key for key in self.parts.keys()}
        return self.parts[key_map[name.lower()]]

    def _get_by_prop(self, value: str, prop: str) -> Part | Plate | Beam | Shape | Material | Pipe | None:
        pmap = {getattr(p, prop): p for p in self.get_all_subparts() + [self]}
        result = pmap.get(value)
        if result is not None:
            return result

        for p in self.get_all_subparts() + [self]:
            for stru_cont in [p.beams, p.plates]:
                res = stru_cont.from_name(value)
                if res is not None:
                    return res

            for shp in p.shapes:
                if getattr(shp, prop) == value:
                    return shp

            for pi in p.pipes:
                if getattr(pi, prop) == value:
                    return pi
                for seg in pi.segments:
                    if getattr(seg, prop) == value:
                        return seg

            for mat in p.materials:
                if getattr(mat, prop) == value:
                    return mat

        logger.debug(f'Unable to find"{value}". Check if the element type is evaluated in the algorithm')
        return None

    def get_by_guid(self, guid) -> Part | Plate | Beam | Shape | Material | Pipe | None:
        return self._get_by_prop(guid, "guid")

    def get_by_name(self, name) -> Part | Plate | Beam | Shape | Material | Pipe | None:
        """Get element of any type by its name."""
        return self._get_by_prop(name, "name")

    def get_all_materials(self, include_self=True) -> list[Material]:
        materials = []
        for part in filter(lambda x: len(x.materials) > 0, self.get_all_parts_in_assembly(include_self=include_self)):
            materials += part.materials.materials

        return materials

    def get_all_sections(self, include_self=True) -> list[Section]:
        sections = []
        for sec in filter(lambda x: len(x.sections) > 0, self.get_all_parts_in_assembly(include_self=include_self)):
            sections += sec.sections.sections
        return sections

    def consolidate_sections(self, include_self=True):
        """Moves all sections from all sub-parts to this part"""
        from ada import Beam
        from ada.fem import FemSection

        new_sections = Sections(parent=self)
        refs_num = 0

        for sec in self.get_all_sections(include_self=include_self):
            res = new_sections.add(sec)
            if res.guid != sec.guid:
                refs = [r for r in sec.refs]
                for elem in refs:
                    refs_num += 1
                    sec.refs.pop(sec.refs.index(elem))
                    if elem not in res.refs:
                        res.refs.append(elem)
                    if isinstance(elem, (Beam, FemSection)):
                        if isinstance(elem, Beam) and sec.guid == elem.taper.guid:
                            elem.taper = res
                        elem.section = res
                    else:
                        raise NotImplementedError(f"Not yet support section {type(elem)=}")

        for part in filter(lambda x: len(x.sections) > 0, self.get_all_parts_in_assembly(include_self=include_self)):
            part.sections = Sections(parent=part)

        sec_map = {sec.guid: sec for sec in new_sections.sections}

        not_found = []
        for beam in self.get_all_physical_objects(by_type=Beam):
            if beam.taper.guid not in sec_map.keys():
                not_found.append((beam, "taper", beam.taper))
            if beam.section.guid not in sec_map.keys():
                not_found.append((beam, "section", beam.section))

        if len(not_found) > 0:
            raise ValueError(f"The following beam sections are not consolidated {not_found}")

        self.sections = new_sections

        return self.sections.sections

    def consolidate_materials(self, include_self=True):
        from ada import Beam, Pipe, PipeSegElbow, PipeSegStraight, Plate
        from ada.fem import FemSection

        num_elem_changed = 0
        new_materials = Materials(parent=self)
        for mat in self.get_all_materials(include_self=include_self):
            res = new_materials.add(mat)
            if res.guid != mat.guid:
                refs = [r for r in mat.refs]
                for elem in refs:
                    mat.refs.pop(mat.refs.index(elem))
                    if elem not in res.refs:
                        res.refs.append(elem)
                    if isinstance(elem, (Beam, Plate, FemSection, PipeSegStraight, PipeSegElbow, Pipe)):
                        elem.material = res
                        num_elem_changed += 1
                    else:
                        raise NotImplementedError(f"Not yet support section {type(elem)=}")

        for part in filter(lambda x: len(x.materials) > 0, self.get_all_parts_in_assembly(include_self=include_self)):
            part.materials = Materials(parent=part)

        self.materials = new_materials

        return self.materials.materials

    def get_all_parts_in_assembly(self, include_self=False, by_type=None) -> list[Part]:
        parent = self.get_assembly()
        list_of_ps = []
        self._flatten_list_of_subparts(parent, list_of_ps)
        if include_self:
            list_of_ps += [self]

        if by_type is not None:
            return list(filter(lambda x: issubclass(type(x), by_type), list_of_ps))

        return list_of_ps

    def get_all_subparts(self, include_self=False) -> list[Part]:
        list_of_parts = [] if include_self is False else [self]
        self._flatten_list_of_subparts(self, list_of_parts)
        return list_of_parts

    def get_all_physical_objects(
        self, sub_elements_only=False, by_type=None, filter_by_guids: list[str] = None, pipe_to_segments=False
    ) -> Iterable[Beam | Plate | Wall | Pipe | Shape]:
        physical_objects = []
        if sub_elements_only:
            iter_parts = iter([self])
        else:
            iter_parts = iter(self.get_all_subparts(include_self=True))

        for p in iter_parts:
            if pipe_to_segments:
                segments = chain.from_iterable([pipe.segments for pipe in p.pipes])
                all_as_iterable = chain(p.plates, p.beams, p.shapes, segments, p.walls)
            else:
                all_as_iterable = chain(p.plates, p.beams, p.shapes, p.pipes, p.walls)
            physical_objects.append(all_as_iterable)

        if by_type is not None:
            res = filter(lambda x: type(x) is by_type, chain.from_iterable(physical_objects))
        else:
            res = chain.from_iterable(physical_objects)

        if filter_by_guids is not None:
            res = filter(lambda x: x.guid in filter_by_guids, res)

        return res

    def beam_clash_check(self, margins=5e-5):
        """
        For all beams in a Assembly get all beams touching or within the beam. Essentially a clash check is performed
        and it returns a dictionary of all beam ids and the touching beams. A margin to the beam volume can be included.

        :param margins: Add margins to the volume box (equal in all directions). Input is in meters. Can be negative.
        :return: A map generator for the list of beams and resulting intersecting beams
        """
        from ada.core.clash_check import basic_intersect

        all_parts = self.get_all_subparts() + [self]
        all_beams = [bm for p in all_parts for bm in p.beams]

        return filter(None, [basic_intersect(bm, margins, all_parts) for bm in all_beams])

    def move_all_mats_and_sec_here_from_subparts(self):
        for p in self.get_all_subparts():
            self._materials += p.materials
            self._sections += p.sections
            p._materials = Materials(parent=p)
            p._sections = Sections(parent=p)

        self.sections.merge_sections_by_properties()
        self.materials.merge_materials_by_properties()

    def _flatten_list_of_subparts(self, p, list_of_parts=None):
        for value in p.parts.values():
            list_of_parts.append(value)
            self._flatten_list_of_subparts(value, list_of_parts)

    def _on_import(self):
        """A method call that will be triggered when a Part is imported into an existing Assembly/Part"""
        raise NotImplementedError()

    def to_fem_obj(
        self,
        mesh_size: float,
        bm_repr: GeomRepr = GeomRepr.LINE,
        pl_repr: GeomRepr = GeomRepr.SHELL,
        shp_repr: GeomRepr = GeomRepr.SOLID,
        options: GmshOptions = None,
        silent=True,
        interactive=False,
        use_quads=False,
        use_hex=False,
        experimental_bm_splitting=True,
    ) -> FEM:
        from ada import Beam, Plate, Shape
        from ada.fem.meshing import GmshOptions, GmshSession

        if isinstance(bm_repr, str):
            bm_repr = GeomRepr.from_str(bm_repr)
        if isinstance(pl_repr, str):
            pl_repr = GeomRepr.from_str(pl_repr)
        if isinstance(shp_repr, str):
            shp_repr = GeomRepr.from_str(shp_repr)

        options = GmshOptions(Mesh_Algorithm=8) if options is None else options
        masses: list[Shape] = []
        with GmshSession(silent=silent, options=options) as gs:
            for obj in self.get_all_physical_objects(sub_elements_only=False):
                if isinstance(obj, Beam):
                    gs.add_obj(obj, geom_repr=bm_repr, build_native_lines=False)
                elif isinstance(obj, Plate):
                    gs.add_obj(obj, geom_repr=pl_repr)
                elif issubclass(type(obj), Shape) and obj.mass is not None:
                    masses.append(obj)
                elif issubclass(type(obj), Shape):
                    gs.add_obj(obj, geom_repr=shp_repr)
                else:
                    logger.error(f'Unsupported object type "{obj}". Should be either plate or beam objects')

            if interactive is True:
                gs.open_gui()

            gs.split_plates_by_beams()

            if experimental_bm_splitting is True and len(list(self.get_all_physical_objects(by_type=Plate))) == 0:
                gs.split_crossing_beams()

            gs.mesh(mesh_size, use_quads=use_quads, use_hex=use_hex)

            if interactive is True:
                gs.open_gui()

            fem = gs.get_fem()

        for mass_shape in masses:
            cog_absolute = mass_shape.placement.absolute_placement() + mass_shape.cog
            n = fem.nodes.add(Node(cog_absolute))
            fem.add_mass(Mass(f"{mass_shape.name}_mass", [n], mass_shape.mass))

        # Move FEM mesh to match part placement origin
        x, y, z = self.placement.origin
        if x != 0.0 or y != 0.0 or z == 0.0:
            fem.nodes.move(self.placement.origin)

        return fem

    def to_vis_mesh(
        self,
        export_config: ExportConfig = None,
        merge_by_color=True,
        opt_func: Callable = None,
        overwrite_cache=False,
        auto_sync_ifc_store=True,
        use_experimental=True,
        cpus: int = None,
    ) -> VisMesh:
        from ada.visualize.interface import part_to_vis_mesh2

        return part_to_vis_mesh2(self, auto_sync_ifc_store, cpus=cpus)

    def to_gltf(
        self,
        gltf_file: str | pathlib.Path,
        auto_sync_ifc_store=True,
        cpus=None,
        limit_to_guids=None,
        embed_meta=False,
        merge_by_color=False,
    ):
        from ada.visualize.interface import part_to_vis_mesh2

        vm = part_to_vis_mesh2(self, auto_sync_ifc_store, cpus=cpus)
        if merge_by_color:
            vm = vm.merge_objects_in_parts_by_color()

        vm.to_gltf(gltf_file, only_these_guids=limit_to_guids, embed_meta=embed_meta)

    def to_stp(
        self,
        destination_file,
        geom_repr: GeomRepr = GeomRepr.SOLID,
        progress_callback: Callable[
            [int, int],
            None,
        ] = None,
    ):
        from ada.occ.store import OCCStore

        step_writer = OCCStore.get_writer()

        num_shapes = len(list(self.get_all_physical_objects()))
        for i, (obj, shape) in enumerate(OCCStore.shape_iterator(self, geom_repr=geom_repr), start=1):
            step_writer.add_shape(shape, obj.name, rgb_color=obj.colour_norm)
            if progress_callback is not None:
                progress_callback(i, num_shapes)

        step_writer.export(destination_file)

    @property
    def parts(self) -> dict[str, Part]:
        return self._parts

    @property
    def shapes(self) -> list[Shape]:
        return self._shapes

    @shapes.setter
    def shapes(self, value: list[Shape]):
        self._shapes = value

    @property
    def beams(self) -> Beams:
        return self._beams

    @beams.setter
    def beams(self, value: Beams):
        self._beams = value

    @property
    def plates(self) -> Plates:
        return self._plates

    @plates.setter
    def plates(self, value: Plates):
        self._plates = value

    @property
    def pipes(self) -> list[Pipe]:
        return self._pipes

    @pipes.setter
    def pipes(self, value: list[Pipe]):
        self._pipes = value

    @property
    def welds(self) -> list[Weld]:
        return self._welds

    @property
    def walls(self) -> list[Wall]:
        return self._walls

    @walls.setter
    def walls(self, value: list[Wall]):
        self._walls = value

    @property
    def nodes(self) -> Nodes:
        return self._nodes

    @property
    def fem(self) -> FEM:
        return self._fem

    @fem.setter
    def fem(self, value: FEM):
        value.parent = self
        self._fem = value

    @property
    def connections(self) -> Connections:
        return self._connections

    @property
    def sections(self) -> Sections:
        return self._sections

    @sections.setter
    def sections(self, value: Sections):
        self._sections = value

    @property
    def materials(self) -> Materials:
        return self._materials

    @materials.setter
    def materials(self, value: Materials):
        self._materials = value

    @property
    def colour(self):
        if self._colour is None:
            from random import randint

            self._colour = randint(0, 255) / 255, randint(0, 255) / 255, randint(0, 255) / 255

        return self._colour

    @colour.setter
    def colour(self, value):
        self._colour = value

    @property
    def properties(self):
        return self._props

    @property
    def placement(self) -> Placement:
        return self._placement

    @placement.setter
    def placement(self, value: Placement):
        self._placement = value

    @property
    def instances(self) -> dict[Any, Instance]:
        return self._instances

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            for bm in self.beams:
                bm.units = value

            for pl in self.plates:
                pl.units = value

            for pipe in self._pipes:
                pipe.units = value

            for shp in self._shapes:
                shp.units = value

            for wall in self.walls:
                wall.units = value

            for pen in self.penetrations:
                pen.units = value

            for p in self.get_all_subparts():
                p.units = value

            self.sections.units = value
            self.materials.units = value
            self._units = value

            if isinstance(self, Assembly):
                from ada.ifc.utils import assembly_to_ifc_file

                self._ifc_file = assembly_to_ifc_file(self)

    @property
    def groups(self) -> dict[str, Group]:
        return self._groups

    @property
    def ifc_class(self) -> SpatialTypes:
        return self._ifc_class

    def __truediv__(self, other_object):
        if type(other_object) in [list, tuple]:
            for obj in other_object:
                self.add_object(obj)
        else:
            self.add_object(other_object)

        return self

    def __repr__(self):
        nbms = len(self.beams) + len([bm for p in self.get_all_subparts() for bm in p.beams])
        npls = len(self.plates) + len([pl for p in self.get_all_subparts() for pl in p.plates])
        npipes = len(self.pipes) + len([pl for p in self.get_all_subparts() for pl in p.pipes])
        nshps = len(self.shapes) + len([shp for p in self.get_all_subparts() for shp in p.shapes])
        nels = len(self.fem.elements) + len([el for p in self.get_all_subparts() for el in p.fem.elements])
        nnodes = len(self.fem.nodes) + len([no for p in self.get_all_subparts() for no in p.fem.nodes])
        return (
            f'Part("{self.name}": Beams: {nbms}, Plates: {npls}, '
            f"Pipes: {npipes}, Shapes: {nshps}, Elements: {nels}, Nodes: {nnodes})"
        )


class Assembly(Part):
    """The Assembly object. A top level container of parts, beams, plates, shapes and FEM."""

    def __init__(
        self,
        name="Ada",
        project="AdaProject",
        user: User = User(),
        schema="IFC4X1",
        settings=Settings(),
        metadata=None,
        units: Units | str = Units.M,
        ifc_settings=None,
        enable_cache: bool = False,
        clear_cache: bool = False,
        ifc_class: SpatialTypes = SpatialTypes.IfcSite,
    ):
        from ada.ifc.store import IfcStore
        from ada.ifc.utils import assembly_to_ifc_file

        metadata = dict() if metadata is None else metadata
        metadata["project"] = project
        metadata["schema"] = schema
        super(Assembly, self).__init__(name=name, settings=settings, metadata=metadata, units=units)
        self.fem.parent = self
        user.parent = self
        self._user = user

        self._ifc_class = ifc_class
        self._ifc_file = assembly_to_ifc_file(self)
        self._convert_options = _ConvertOptions()
        self._ifc_sections = None
        self._ifc_materials = None
        self._source_ifc_files = dict()
        self._ifc_settings = ifc_settings
        self._presentation_layers = PresentationLayers()

        self._ifc_store = IfcStore(assembly=self)
        self._cache_store = None
        if enable_cache:
            self._cache_store = CacheStore(name)
            self.cache_store.sync(self, clear_cache=clear_cache)

    def read_ifc(
        self, ifc_file: str | os.PathLike | ifcopenshell.file, data_only=False, elements2part=None, create_cache=False
    ):
        """Import from IFC file."""

        if self.cache_store is not None and isinstance(ifc_file, ifcopenshell.file) is False:
            if self.cache_store.from_cache(self, ifc_file) is True:
                return None

        self.ifc_store.load_ifc_content_from_file(ifc_file, data_only=data_only, elements2part=elements2part)

        if self.cache_store is not None:
            self.cache_store.to_cache(self, ifc_file, create_cache)

    def read_fem(
        self,
        fem_file: str | os.PathLike,
        fem_format: FEATypes | str = None,
        name: str = None,
        fem_converter: FemConverters | str = "default",
        cache_model_now=False,
    ):
        """Import a Finite Element model. Currently supported FEM formats: Abaqus, Sesam and Calculix"""
        from ada.fem.formats.general import get_fem_converters

        fem_file = pathlib.Path(fem_file)
        if fem_file.exists() is False:
            raise FileNotFoundError(fem_file)

        if self.cache_store is not None:
            if self.cache_store.from_cache(self, fem_file) is True:
                return None

        fem_importer, _ = get_fem_converters(fem_file, fem_format, fem_converter)
        if fem_importer is None:
            suffix = fem_file.suffix
            raise FormatNotSupportedException(f'File "{fem_file.name}" [{suffix}] is not a supported FEM format.')

        temp_assembly: Assembly = fem_importer(fem_file, name)
        self.__add__(temp_assembly)

        if self.cache_store is not None:
            self.cache_store.to_cache(self, fem_file, cache_model_now)

    def to_fem(
        self,
        name: str,
        fem_format: FEATypes | str,
        scratch_dir=None,
        metadata=None,
        execute=False,
        run_ext=False,
        cpus=2,
        gpus=None,
        overwrite=False,
        fem_converter="default",
        exit_on_complete=True,
        run_in_shell=False,
        make_zip_file=False,
        return_fea_results=True,
    ) -> FEAResult | None:
        """
        Create a FEM input file deck for executing fem analysis in a specified FEM format.
        Currently there is limited write support for the following FEM formats:

        Open Source

        * Calculix
        * Code_Aster

        not open source

        * Abaqus
        * Usfos
        * Sesam


        Write support is added on a need-only-basis. Any contributions are welcomed!

        :param name: Name of FEM analysis input deck
        :param fem_format: Desired fem format
        :param scratch_dir: Output directory for analysis input deck
        :param metadata: Parse additional commands to FEM solver not supported by the generalized classes
        :param execute: Execute analysis on complete
        :param run_ext: Run analysis externally or wait for complete
        :param cpus: Number of cpus for running the analysis
        :param gpus: Number of gpus for running the analysis (wherever relevant)
        :param overwrite: Overwrite existing input file deck
        :param fem_converter: Set desired fem converter. Use either 'default' or 'meshio'.
        :param exit_on_complete:
        :param run_in_shell:
        :param make_zip_file:
        :param return_fea_results: Automatically import the result mesh into

            Note! Meshio implementation currently only supports reading & writing elements and nodes.

        Abaqus Metadata:

            'ecc_to_mpc': Runs the method :func:`~ada.fem.FEM.convert_ecc_to_mpc` . Default is True
            'hinges_to_coupling': Runs the method :func:`~ada.fem.FEM.convert_hinges_2_couplings` . Default is True

            Important Note! The ecc_to_mpc and hinges_to_coupling will make permanent modifications to the model.
            If this proves to create issues regarding performance this should be evaluated further.

        """
        from ada.fem.formats.execute import execute_fem
        from ada.fem.formats.general import FEATypes, write_to_fem
        from ada.fem.formats.postprocess import postprocess
        from ada.fem.formats.utils import default_fem_res_path

        if isinstance(fem_format, str):
            fem_format = FEATypes.from_str(fem_format)

        scratch_dir = Settings.scratch_dir if scratch_dir is None else pathlib.Path(scratch_dir)

        write_to_fem(self, name, fem_format, overwrite, fem_converter, scratch_dir, metadata, make_zip_file)

        if execute:
            execute_fem(
                name, fem_format, scratch_dir, cpus, gpus, run_ext, metadata, execute, exit_on_complete, run_in_shell
            )

        fem_res_files = default_fem_res_path(name, scratch_dir=scratch_dir)
        res_path = fem_res_files.get(fem_format, None)

        if res_path.exists() is False or return_fea_results is False:
            return None

        return postprocess(res_path, fem_format=fem_format)

    def to_ifc(
        self,
        destination=None,
        include_fem=False,
        file_obj_only=False,
        validate=False,
        progress_callback: Callable[[int, int], None] = None,
    ) -> ifcopenshell.file:
        import ifcopenshell.validate

        if destination is None or file_obj_only is True:
            destination = "object"
        else:
            destination = pathlib.Path(destination).resolve().absolute()

        print(f'Beginning writing to IFC file "{destination}" using IfcOpenShell')

        self.ifc_store.sync(include_fem=include_fem, progress_callback=progress_callback)

        if file_obj_only is False:
            os.makedirs(destination.parent, exist_ok=True)
            self.ifc_store.save_to_file(destination)

        if validate:
            ifcopenshell.validate.validate(self.ifc_store.f, logging)

        print("IFC file creation complete")
        return self.ifc_store.f

    def to_genie_xml(self, destination_xml):
        from ada.fem.formats.sesam.xml.write.write_xml import write_xml

        write_xml(self, destination_xml)

    def push(self, comment, bimserver_url, username, password, project, merge=False, sync=False):
        """Push current assembly to BimServer with a comment tag that defines the revision name"""
        from ada.core.bimserver import BimServerConnect

        bimcon = BimServerConnect(bimserver_url, username, password, self)
        bimcon.push(project, comment, merge, sync)

    def pull(self, bimserver_url, username, password, project, checkout=False):
        from ada.core.bimserver import BimServerConnect

        bimcon = BimServerConnect(bimserver_url, username, password, self)
        bimcon.pull(project, checkout)

    def get_ifc_source_by_name(self, ifc_file):
        from ada.ifc.read.reader_utils import open_ifc

        if ifc_file not in self._source_ifc_files.keys():
            ifc_f = open_ifc(ifc_file)
            self._source_ifc_files[ifc_file] = ifc_f
        else:
            ifc_f = self._source_ifc_files[ifc_file]

        return ifc_f

    @property
    def ifc_store(self) -> IfcStore:
        return self._ifc_store

    @ifc_store.setter
    def ifc_store(self, value):
        self._ifc_store = value

    @property
    def presentation_layers(self) -> PresentationLayers:
        return self._presentation_layers

    @presentation_layers.setter
    def presentation_layers(self, value):
        self._presentation_layers = value

    @property
    def user(self) -> User:
        return self._user

    @property
    def convert_options(self) -> _ConvertOptions:
        return self._convert_options

    @property
    def cache_store(self) -> CacheStore:
        return self._cache_store

    def __add__(self, other: Assembly | Part):
        if other.units != self.units:
            other.units = self.units

        for interface_n in other.fem.interface_nodes:
            n = interface_n.node
            for p in self.get_all_parts_in_assembly(True):
                res = p.fem.nodes.get_by_volume(n.p)
                if res is not None and len(res) > 0:
                    replace_node = res[0]
                    for ref in n.refs:
                        if isinstance(ref, Connector):
                            if n == ref.n1:
                                ref.n1 = replace_node
                            elif n == ref.n2:
                                ref.n2 = replace_node
                            else:
                                logger.warning(f'No matching node found for either n1 or n2 of "{ref}"')
                        elif isinstance(ref, Csys):
                            index = ref.nodes.index(n)
                            ref.nodes.pop(index)
                            ref.nodes.insert(index, replace_node)
                        elif isinstance(ref, FemSet):
                            index = ref.members.index(n)
                            ref.members.pop(index)
                            ref.members.insert(index, replace_node)
                        else:
                            raise NotImplementedError(f'Unsupported type "{type(ref)}"')
                    break

        self.fem += other.fem

        for p in other.parts.values():
            p.parent = self
            self.add_part(p)

        for mat in other.materials:
            if mat not in self.materials:
                self.materials.add(mat)

        self.sections += other.sections
        self.shapes += other.shapes
        self.beams += other.beams
        self.plates += other.plates
        self.pipes += other.pipes
        self.walls += other.walls
        return self

    def __repr__(self):
        nbms = len([bm for p in self.get_all_subparts() for bm in p.beams]) + len(self.beams)
        npls = len([pl for p in self.get_all_subparts() for pl in p.plates]) + len(self.plates)
        nshps = len([shp for p in self.get_all_subparts() for shp in p.shapes]) + len(self.shapes)
        npipes = len(self.pipes) + len([pl for p in self.get_all_subparts() for pl in p.pipes])
        nels = len(self.fem.elements) + len([el for p in self.get_all_subparts() for el in p.fem.elements])
        nns = len(self.fem.nodes) + len([no for p in self.get_all_subparts() for no in p.fem.nodes])
        return (
            f'Assembly("{self.name}": Beams: {nbms}, Plates: {npls}, Pipes: {npipes}, '
            f"Shapes: {nshps}, Elements: {nels}, Nodes: {nns})"
        )



from __future__ import annotations

import subprocess
from typing import TYPE_CHECKING, Callable

from ada.base.types import BaseEnum
from ada.config import get_logger

from . import abaqus, calculix, code_aster, sesam, usfos
from .utils import interpret_fem_format_from_path

if TYPE_CHECKING:
    from ada import Assembly

logger = get_logger()


class FEATypes(BaseEnum):
    CODE_ASTER = "code_aster"
    CALCULIX = "calculix"
    ABAQUS = "abaqus"
    SESAM = "sesam"
    USFOS = "usfos"
    GMSH = "gmsh"

    # formats only
    XDMF = "xdmf"

    @staticmethod
    def get_solvers_only():
        non_solvers = [FEATypes.XDMF, FEATypes.GMSH]
        return [x for x in FEATypes if x not in non_solvers]


fem_imports: dict[FEATypes, Callable[..., Assembly]] = {
    FEATypes.ABAQUS: abaqus.read_fem,
    FEATypes.SESAM: sesam.read_fem,
    FEATypes.CODE_ASTER: code_aster.read_fem,
}

fem_exports = {
    FEATypes.ABAQUS: abaqus.to_fem,
    FEATypes.CALCULIX: calculix.to_fem,
    FEATypes.CODE_ASTER: code_aster.to_fem,
    FEATypes.SESAM: sesam.to_fem,
    FEATypes.USFOS: usfos.to_fem,
}

fem_executables: dict[FEATypes, Callable[..., subprocess.CompletedProcess]] = {
    FEATypes.ABAQUS: abaqus.run_abaqus,
    FEATypes.CALCULIX: calculix.run_calculix,
    FEATypes.CODE_ASTER: code_aster.run_code_aster,
    FEATypes.SESAM: sesam.run_sesam,
}

fem_solver_map = {FEATypes.SESAM: "sestra", FEATypes.CALCULIX: "ccx"}


class FemConverters(BaseEnum):
    DEFAULT = "default"
    MESHIO = "meshio"


def get_fem_converters(fem_file, fem_format: str | FEATypes, fem_converter: str | FemConverters):
    from ada.fem.formats.mesh_io import meshio_read_fem, meshio_to_fem

    if isinstance(fem_format, str):
        fem_format = FEATypes.from_str(fem_format)
    if isinstance(fem_converter, str):
        fem_converter = FemConverters.from_str(fem_converter)

    if fem_format is None:
        fem_format = interpret_fem_format_from_path(fem_file)

    if fem_converter == FemConverters.DEFAULT:
        fem_importer = fem_imports.get(fem_format, None)
        fem_exporter = fem_exports.get(fem_format, None)
    elif fem_converter == FemConverters.MESHIO:
        fem_importer = meshio_read_fem
        fem_exporter = meshio_to_fem
    else:
        raise ValueError(f'Unrecognized fem_converter "{fem_converter}". Only "meshio" and "default" are supported')

    return fem_importer, fem_exporter


def export_fem(assembly, name, analysis_dir, fem_format, fem_converter, metadata):
    _, fem_exporter = get_fem_converters("", fem_format, fem_converter)
    metadata = dict() if metadata is None else metadata
    metadata["fem_format"] = fem_format
    try:
        fem_exporter(assembly, name, analysis_dir, metadata)
        return True
    except IOError as e:
        logger.error(e)
        return False


def write_to_fem(
    assembly: Assembly,
    name: str,
    fem_format: FEATypes,
    overwrite: bool,
    fem_converter: str,
    scratch_dir,
    metadata: dict,
    make_zip_file,
):
    from ada.fem.formats.utils import default_fem_res_path, folder_prep, should_convert

    fem_res_files = default_fem_res_path(name, scratch_dir=scratch_dir)

    res_path = fem_res_files.get(fem_format, None)
    metadata = dict() if metadata is None else metadata
    metadata["fem_format"] = fem_format.value

    out = None
    if should_convert(res_path, overwrite):
        analysis_dir = folder_prep(scratch_dir, name, overwrite)
        _, fem_exporter = get_fem_converters("", fem_format, fem_converter)

        if fem_exporter is None:
            raise ValueError(f'FEM export for "{fem_format}" using "{fem_converter}" is currently not supported')

        fem_exporter(assembly, name, analysis_dir, metadata)

        if make_zip_file is True:
            import shutil

            shutil.make_archive(name, "zip", str(analysis_dir))
    else:
        print(f'Result file "{res_path}" already exists.\nUse "overwrite=True" if you wish to overwrite')

    if out is None and res_path is None:
        logger.info("No Result file is created")
        return None



import base64
import xml.etree.ElementTree as ET
import zipfile
from io import BytesIO

from ada.config import get_logger

logger = get_logger()


def xml_elem_to_sat_text(sat_el: ET.Element) -> str:
    if sat_el.tag == "sat_embedded":
        text = sat_el.text
        data = base64.b64decode(text)
    elif sat_el.tag == "sat_embedded_sequence":
        data = b""
        for res in sat_el.findall(".//cdata_segment"):
            res = base64.b64decode(res.text)
            data += res
    else:
        raise NotImplementedError(f'SAT el Tag type "{sat_el.tag}" is not yet added')

    byio = BytesIO(data)
    try:
        zipdata = zipfile.ZipFile(byio)
    except zipfile.BadZipFile:
        return str(data, encoding="utf-8")
    res = {name: zipdata.read(name) for name in zipdata.namelist()}
    if len(res.keys()) != 1:
        raise NotImplementedError("No support for binary zip data containing multipart SAT file yet")

    return str(res["b64temp.sat"], encoding="utf-8").replace("\r", "")


def write_xml_sat_text_to_file(xml_file, out_file):
    xml_root = ET.parse(str(xml_file)).getroot()
    with open(out_file, "w") as f:
        for sat_geometry_el in xml_root.iterfind(".//sat_embedded"):
            f.write(xml_elem_to_sat_text(sat_geometry_el))
        for sat_geometry_seq in xml_root.iterfind(".//sat_embedded_sequence"):
            f.write(xml_elem_to_sat_text(sat_geometry_seq))


def get_sat_text_from_xml(xml_file):
    xml_root = ET.parse(str(xml_file)).getroot()
    sat_text = ""

    for sat_geometry_el in xml_root.findall(".//sat_embedded"):
        sat_text += xml_elem_to_sat_text(sat_geometry_el)

    for sat_geometry_seq in xml_root.findall(".//sat_embedded_sequence"):
        sat_text += xml_elem_to_sat_text(sat_geometry_seq)

    return sat_text.replace("\r", "")



from __future__ import annotations

from dataclasses import dataclass

import ifcopenshell
import numpy as np

from ada import Section
from ada.config import Settings, get_logger
from ada.sections.categories import SectionCat

from ..utils import create_ifcindexpolyline, create_ifcpolyline

logger = get_logger()


class UnrecognizedSectionType(Exception):
    pass


def get_profile_class(section: Section) -> ProfileBase:
    st = Section.TYPES

    if section.type == st.IPROFILE:
        return IProfile(section)
    elif section.type == st.ANGULAR:
        return AngularProfile(section)
    elif section.type == st.BOX:
        return BoxProfile(section)
    elif section.type == st.TPROFILE:
        return TProfile(section)
    elif section.type == st.CHANNEL:
        return ChannelProfile(section)
    elif section.type == st.CIRCULAR:
        return CircularProfile(section)
    elif section.type == st.TUBULAR:
        return TubularProfile(section)
    elif section.type == st.GENERAL:
        return GeneralProfile(section)
    elif section.type == st.FLATBAR:
        return FlatBarProfile(section)
    elif section.type == st.POLY:
        return PolyProfile(section)
    else:
        raise UnrecognizedSectionType(f"Type -> {section.type}")


def export_beam_section_profile_def(section: Section):
    if section.parent is None or section.parent.parent is None:
        raise ValueError("Lacking parent")

    if section.name is None:
        raise ValueError("Name cannot be None!")

    a = section.parent.parent.get_assembly()
    f = a.ifc_store.f

    sec_props = dict(ProfileType="AREA", ProfileName=section.name)

    section_profile_instance = get_profile_class(section)

    sec_props_input = section_profile_instance.get_ifc_props(f)
    sec_props.update(sec_props_input)

    ifc_sec_type = section_profile_instance.get_ifc_type()

    profile = f.create_entity(ifc_sec_type, **sec_props)

    return profile


@dataclass
class ProfileBase:
    section: Section

    def get_ifc_type(self) -> str:
        ...

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        ...


@dataclass
class IProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        if Settings.force_param_profiles is False:
            return "IfcArbitraryClosedProfileDef"
        else:
            return "IfcIShapeProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        if Settings.force_param_profiles is False:
            section_profile = section.get_section_profile(True)
            polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)

            sec_props = dict(OuterCurve=polyline)
        else:
            if SectionCat.is_strong_axis_symmetric(section) is False:
                logger.info("Note! IfcAsymmetricIShapeProfileDef as it is not supported by ifcopenshell v IFC4")
            sec_props = dict(
                OverallWidth=section.w_top,
                OverallDepth=section.h,
                WebThickness=section.t_w,
                FlangeThickness=section.t_ftop,
            )
        return sec_props


@dataclass
class TProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        if Settings.force_param_profiles is False:
            return "IfcArbitraryClosedProfileDef"
        else:
            return "IfcTShapeProfileDef"

    def get_ifc_props(self, f) -> dict:
        section = self.section
        if Settings.force_param_profiles is False:
            section_profile = section.get_section_profile(True)
            polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)

            sec_props = dict(OuterCurve=polyline)
        else:
            if SectionCat.is_strong_axis_symmetric(section) is False:
                logger.info(
                    "Note! Not using IfcAsymmetricIShapeProfileDef as it is not supported by ifcopenshell v IFC4"
                )
            sec_props = dict(
                FlangeWidth=section.w_top,
                Depth=section.h,
                WebThickness=section.t_w,
                FlangeThickness=section.t_ftop,
            )
        return sec_props


@dataclass
class AngularProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        return "IfcArbitraryClosedProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        if Settings.force_param_profiles is True:
            logger.debug(f'Export of "{section.type}" profile to parametric IFC profile is not yet added')
        section_profile = section.get_section_profile(True)
        points = [f.createIfcCartesianPoint(p) for p in section_profile.outer_curve.points2d]
        ifc_polyline = f.createIfcPolyLine(points)
        return dict(OuterCurve=ifc_polyline)


@dataclass
class BoxProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        return "IfcArbitraryProfileDefWithVoids"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        if Settings.force_param_profiles is True:
            logger.debug(f'Export of "{section.type}" profile to parametric IFC profile is not yet added')
        section_profile = section.get_section_profile(True)
        ot_disc = section_profile.outer_curve.points2d
        in_disc = section_profile.inner_curve.points2d
        outer_points = [f.createIfcCartesianPoint(p) for p in ot_disc + [ot_disc[0]]]
        inner_points = [f.createIfcCartesianPoint(p) for p in in_disc + [in_disc[0]]]
        inner_curve = f.createIfcPolyLine(inner_points)
        outer_curve = f.createIfcPolyLine(outer_points)
        return dict(OuterCurve=outer_curve, InnerCurves=[inner_curve])


@dataclass
class CircularProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        return "IfcCircleProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        return dict(Radius=self.section.r)


@dataclass
class TubularProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        return "IfcCircleHollowProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        return dict(Radius=section.r, WallThickness=section.wt)


@dataclass
class GeneralProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        return "IfcCircleProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        logger.warning("Note! Creating a Circle profile from general section (just for visual inspection as of now)")
        r = np.sqrt(self.section.properties.Ax / np.pi)
        return dict(Radius=r)


@dataclass
class FlatBarProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        return "IfcArbitraryClosedProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        if Settings.force_param_profiles is True:
            logger.debug(f'Export of "{section.type}" profile to parametric IFC profile is not yet added')
        section_profile = section.get_section_profile(True)
        polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)
        return dict(OuterCurve=polyline)


@dataclass
class ChannelProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        if Settings.force_param_profiles is False:
            return "IfcArbitraryClosedProfileDef"
        else:
            return "IfcUShapeProfileDef"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        if Settings.force_param_profiles is False:
            section_profile = section.get_section_profile(True)
            polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)

            props = dict(OuterCurve=polyline)
        else:
            props = dict(
                Depth=section.h, FlangeWidth=section.w_top, WebThickness=section.t_w, FlangeThickness=section.t_ftop
            )
        return props


@dataclass
class PolyProfile(ProfileBase):
    def get_ifc_type(self) -> str:
        if self.section.poly_inner is None:
            return "IfcArbitraryClosedProfileDef"
        else:
            return "IfcArbitraryProfileDefWithVoids"

    def get_ifc_props(self, f: ifcopenshell.file) -> dict:
        section = self.section
        opoly = section.poly_outer
        opoints = [(float(n[0]), float(n[1]), float(n[2])) for n in opoly.seg_global_points]
        opolyline = create_ifcindexpolyline(f, opoints, opoly.seg_index)
        if section.poly_inner is None:
            props = dict(OuterCurve=opolyline)
        else:
            ipoly = section.poly_inner
            ipoints = [(float(n[0]), float(n[1]), float(n[2])) for n in ipoly.seg_global_points]
            ipolyline = create_ifcindexpolyline(f, ipoints, ipoly.seg_index)

            props = dict(OuterCurve=opolyline, InnerCurves=[ipolyline])

        return props



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from ada.concepts.points import Node

from .common import FemBase

if TYPE_CHECKING:
    from .elements import Elem


class SetTypes:
    NSET = "nset"
    ELSET = "elset"

    all = [NSET, ELSET]


class FemSet(FemBase):
    """

    :param name: Name of Set
    :param members: Set Members
    :param set_type: Type of set (either 'nset' or 'elset')
    :param metadata: Metadata for object
    :param parent: Parent object
    """

    TYPES = SetTypes

    def __init__(self, name, members: None | list[Elem | Node], set_type=None, metadata=None, parent=None):
        super().__init__(name, metadata, parent)
        from ada.fem import Elem

        if set_type is None:
            set_type = eval_set_type_from_members(members)

        if members is None:
            members = []

        for m in members:
            if isinstance(m, (Elem, Node)):
                m.refs.append(self)

        self._set_type = set_type
        if self.type not in SetTypes.all:
            raise ValueError(f'set type "{set_type}" is not valid')
        self._members = members
        self._refs = []

    def __len__(self):
        return len(self._members)

    def __contains__(self, item):
        return item.id in self._members

    def __getitem__(self, index):
        return self._members[index]

    def __add__(self, other: FemSet) -> FemSet:
        self.add_members(other.members)
        return self

    def add_members(self, members: List[Union[Elem, Node]]):
        self._members += members

    @property
    def type(self):
        return self._set_type.lower()

    @property
    def members(self) -> list[Elem | Node]:
        return self._members

    @property
    def refs(self):
        return self._refs

    def __repr__(self):
        return f'FemSet({self.name}, type: "{self.type}", members: "{len(self.members)}")'


def eval_set_type_from_members(members: list[Elem | Node]) -> str:
    from ada.fem import Elem

    res = set([type(mem) for mem in members])
    if len(res) == 1 and type(members[0]) is Node:
        return FemSet.TYPES.NSET
    elif len(res) == 1 and issubclass(type(members[0]), Elem):
        return FemSet.TYPES.ELSET
    elif len(res) == 1 and type(members[0]) is tuple:
        return FemSet.TYPES.NSET
    else:
        raise ValueError("Currently Mixed Femsets are not allowed")
        # return "mixed"


def is_lazy(members: list[Elem | Node]) -> bool:
    res = set([type(mem) for mem in members])
    if len(res) == 1 and type(members[0]) is tuple:
        return True
    else:
        return False



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

import numpy as np

from ada.base.physical_objects import BackendGeom
from ada.base.units import Units
from ada.config import Settings as _Settings
from ada.config import get_logger
from ada.core.utils import Counter, roundoff
from ada.core.vector_utils import angle_between, calc_zvec, unit_vector, vector_length
from ada.materials.utils import get_material
from ada.sections.utils import get_section

from .curves import ArcSegment
from .points import Node

if TYPE_CHECKING:
    from ada import Section

logger = get_logger()


class Pipe(BackendGeom):
    def __init__(
        self, name, points, sec, mat="S355", content=None, metadata=None, colour=None, units: Units = Units.M, guid=None
    ):
        super().__init__(name, guid=guid, metadata=metadata, units=units)

        self._section, _ = get_section(sec)
        self._section.parent = self
        self._material = get_material(mat)
        self._content = content
        self.colour = colour

        self.section.refs.append(self)
        self.material.refs.append(self)

        self._n1 = points[0] if type(points[0]) is Node else Node(points[0], units=units)
        self._n2 = points[-1] if type(points[-1]) is Node else Node(points[-1], units=units)
        self._points = [Node(n, units=units) if type(n) is not Node else n for n in points]
        self._segments = build_pipe_segments(self)

    @property
    def segments(self) -> list[PipeSegStraight | PipeSegElbow]:
        return self._segments

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def points(self):
        return self._points

    @property
    def start(self):
        return self.points[0]

    @property
    def end(self):
        return self.points[-1]

    @property
    def metadata(self):
        return self._metadata

    @property
    def pipe_bend_radius(self):
        wt = self.section.wt
        r = self.section.r
        d = r * 2
        w_tol = 0.125 if self.units == Units.M else 125
        cor_tol = 0.003 if self.units == Units.M else 3
        corr_t = (wt - (wt * w_tol)) - cor_tol
        d -= 2.0 * corr_t

        return roundoff(d + corr_t / 2.0)

    @property
    def section(self) -> Section:
        return self._section

    @section.setter
    def section(self, value):
        self._section = value

    @property
    def n1(self) -> Node:
        return self._n1

    @property
    def n2(self) -> Node:
        return self._n2

    @property
    def nodes(self) -> list[Node]:
        return [self.n1, self.n2]

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            self.n1.units = value
            self.n2.units = value
            self.section.units = value
            self.material.units = value
            for p in self.points:
                p.units = value
            self._segments = build_pipe_segments(self)
            self._units = value

    def __repr__(self):
        return f"Pipe({self.name}, {self.section})"

    @staticmethod
    def from_segments(name: str, segments: list[PipeSegStraight | PipeSegElbow]) -> Pipe:
        points = list(chain.from_iterable([(Node(x.p1), Node(x.p2)) for x in segments]))
        seg0 = segments[0]
        seg0_section = seg0.section
        seg0_material = seg0.material

        pipe = Pipe(name, points, seg0_section, seg0_material)
        for i, seg in enumerate(segments):
            pipe.segments[i].guid = seg.guid

        return pipe


class PipeSegStraight(BackendGeom):
    def __init__(
        self, name, p1, p2, section, material, parent=None, guid=None, metadata=None, units=Units.M, colour=None
    ):
        super(PipeSegStraight, self).__init__(
            name=name, guid=guid, metadata=metadata, units=units, parent=parent, colour=colour
        )
        self.p1 = p1
        self.p2 = p2
        self.section = section
        self.material = material
        section.refs.append(self)
        material.refs.append(self)

    @property
    def xvec1(self):
        return self.p2.p - self.p1.p

    @property
    def zvec(self):
        return calc_zvec(self.xvec1)

    @property
    def line(self):
        from ada.occ.utils import make_edge

        return make_edge(self.p1, self.p2)

    def shell(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import sweep_pipe

        return sweep_pipe(self.line, self.xvec1, self.section.r, self.section.wt, ElemType.SHELL)

    def solid(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import apply_penetrations, sweep_pipe

        raw_geom = sweep_pipe(self.line, self.xvec1, self.section.r, self.section.wt, ElemType.SOLID)

        geom = apply_penetrations(raw_geom, self.penetrations)
        return geom

    def __repr__(self):
        return f"PipeSegStraight({self.name}, p1={self.p1}, p2={self.p2})"


class PipeSegElbow(BackendGeom):
    def __init__(
        self,
        name,
        p1,
        p2,
        p3,
        bend_radius,
        section,
        material,
        parent=None,
        guid=None,
        metadata=None,
        units=Units.M,
        colour=None,
        arc_seg=None,
    ):
        super(PipeSegElbow, self).__init__(
            name=name, guid=guid, metadata=metadata, units=units, parent=parent, colour=colour
        )
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.bend_radius = bend_radius
        self.section = section
        self.material = material
        self._arc_seg = arc_seg
        section.refs.append(self)
        material.refs.append(self)

    @property
    def parent(self) -> Pipe:
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def xvec1(self):
        return self.p2.p - self.p1.p

    @property
    def xvec2(self):
        return self.p3.p - self.p2.p

    @property
    def zvec(self):
        return calc_zvec(self.xvec1)

    def line(self):
        from ada.core.curve_utils import make_edges_and_fillet_from_3points

        if self.arc_seg.edge_geom is None:
            _, _, fillet = make_edges_and_fillet_from_3points(self.p1, self.p2, self.p3, self.bend_radius)
            edge = fillet
        else:
            edge = self.arc_seg.edge_geom
        return edge

    def shell(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import sweep_pipe

        i = self.parent.segments.index(self)
        if i != 0:
            pseg = self.parent.segments[i - 1]
            xvec = pseg.xvec1
        else:
            xvec = self.xvec1

        return sweep_pipe(self.line(), xvec, self.section.r, self.section.wt, ElemType.SHELL)

    def solid(self):
        from ada.fem.shapes import ElemType
        from ada.occ.utils import apply_penetrations, sweep_pipe

        i = self.parent.segments.index(self)
        if i != 0:
            pseg = self.parent.segments[i - 1]
            xvec = pseg.xvec1
        else:
            xvec = self.xvec1
        raw_geom = sweep_pipe(self.line(), xvec, self.section.r, self.section.wt, ElemType.SOLID)

        geom = apply_penetrations(raw_geom, self.penetrations)
        return geom

    @property
    def arc_seg(self) -> ArcSegment:
        return self._arc_seg

    def __repr__(self):
        return f"PipeSegElbow({self.name}, r={self.bend_radius}, p1={self.p1}, p2={self.p2}, p3={self.p3})"


def build_pipe_segments(pipe: Pipe) -> list[PipeSegStraight, PipeSegElbow]:
    from ada.core.curve_utils import make_arc_segment

    segs = []
    for p1, p2 in zip(pipe.points[:-1], pipe.points[1:]):
        if vector_length(p2.p - p1.p) == 0.0:
            logger.info("skipping zero length segment")
            continue
        segs.append([p1, p2])
    segments = segs

    seg_names = Counter(prefix=pipe.name + "_")

    # Make elbows and adjust segments
    props = dict(section=pipe.section, material=pipe.material, parent=pipe, units=pipe.units)
    angle_tol = 1e-1

    len_tol = _Settings.point_tol if pipe.units == Units.M else _Settings.point_tol * 1000

    pipe_segments = []
    if len(segments) == 1:
        seg_s, seg_e = segments[0]
        pipe_segments.append(PipeSegStraight(next(seg_names), seg_s, seg_e, **props))

    for i, (seg1, seg2) in enumerate(zip(segments[:-1], segments[1:])):
        p11, p12 = seg1
        p21, p22 = seg2
        vlen1 = vector_length(seg1[1].p - seg1[0].p)
        vlen2 = vector_length(seg2[1].p - seg2[0].p)

        if vlen1 < len_tol or vlen2 == len_tol:
            logger.error(f'Segment Length is below point tolerance for unit "{pipe.units}". Skipping')
            continue

        xvec1 = unit_vector(p12.p - p11.p)
        xvec2 = unit_vector(p22.p - p21.p)
        a = angle_between(xvec1, xvec2)
        res = True if abs(abs(a) - abs(np.pi)) < angle_tol or abs(abs(a) - 0.0) < angle_tol else False

        if res is True:
            pipe_segments.append(PipeSegStraight(next(seg_names), p11, p12, **props))
        else:
            if p12 != p21:
                logger.error("No shared point found")

            if i != 0 and len(pipe_segments) > 0:
                pseg = pipe_segments[-1]
                prev_p = (pseg.p1.p, pseg.p2.p)
            else:
                prev_p = (p11.p, p12.p)
            try:
                seg1, arc, seg2 = make_arc_segment(prev_p[0], prev_p[1], p22.p, pipe.pipe_bend_radius * 0.99)
            except ValueError as e:
                logger.error(f"Error: {e}")  # , traceback: "{traceback.format_exc()}"')
                continue
            except RuntimeError as e:
                logger.error(f"Error: {e}")  # , traceback: "{traceback.format_exc()}"')
                continue

            if i == 0 or len(pipe_segments) == 0:
                pipe_segments.append(
                    PipeSegStraight(
                        next(seg_names), Node(seg1.p1, units=pipe.units), Node(seg1.p2, units=pipe.units), **props
                    )
                )
            else:
                if len(pipe_segments) == 0:
                    continue
                pseg = pipe_segments[-1]
                pseg.p2 = Node(seg1.p2, units=pipe.units)

            pipe_segments.append(
                PipeSegElbow(
                    next(seg_names) + "_Elbow",
                    Node(seg1.p1, units=pipe.units),
                    Node(p21.p, units=pipe.units),
                    Node(seg2.p2, units=pipe.units),
                    arc.radius,
                    **props,
                    arc_seg=arc,
                )
            )
            pipe_segments.append(
                PipeSegStraight(
                    next(seg_names), Node(seg2.p1, units=pipe.units), Node(seg2.p2, units=pipe.units), **props
                )
            )

    return pipe_segments



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada.fem import Bc


def create_bc_str(bc: Bc) -> str:
    from ada.fem.utils import is_parent_of_node_solid

    set_name = bc.fem_set.name
    is_solid = False
    for no in bc.fem_set.members:
        is_solid = is_parent_of_node_solid(no)
        if is_solid:
            break
    dofs = ["DX", "DY", "DZ"]
    if is_solid is False:
        dofs += ["DRX", "DRY", "DRZ"]
    bc_str = ""
    for i, n in enumerate(dofs, start=1):
        if i in bc.dofs:
            bc_str += f"{n}=0, "
    dofs_str = f"""dofs = dict(
    GROUP_NO="{set_name}",
    {bc_str}
)\n"""

    return (
        dofs_str
        + f"""{bc.name} = AFFE_CHAR_MECA(
    MODELE=model, DDL_IMPO=_F(**dofs)
)"""
    )



from __future__ import annotations

from ada.fem.shapes.definitions import (
    ConnectorTypes,
    LineShapes,
    MassTypes,
    ShellShapes,
    SolidShapes,
)

sh = ShellShapes
so = SolidShapes
li = LineShapes


ada_to_abaqus_format = {
    sh.TRI: ("S3", "S3R", "R3D3", "S3RS"),
    sh.TRI6: ("STRI65",),
    sh.TRI7: ("S7",),
    sh.QUAD: ("S4", "S4R", "R3D4"),
    sh.QUAD8: ("S8", "S8R"),
    so.HEX8: ("C3D8", "C3D8R", "C3D8H"),
    so.HEX20: ("C3D20", "C3D20R", "C3D20RH"),
    so.HEX27: ("C3D27",),
    so.TETRA: ("C3D4",),
    so.TETRA10: ("C3D10",),
    so.PYRAMID5: ("C3D5", "C3D5H"),
    so.WEDGE: ("C3D6",),
    so.WEDGE15: ("C3D15",),
    li.LINE: ("B31", "B31H"),
    li.LINE3: ("B32",),
    MassTypes.MASS: ("MASS",),
    MassTypes.ROTARYI: ("ROTARYI",),
    ConnectorTypes.CONNECTOR: ("CONN3D2",),
}


def abaqus_el_type_to_ada(el_type):
    for key, val in ada_to_abaqus_format.items():
        if el_type in val:
            return key
    raise ValueError(f'Element type "{el_type}" has not been added to conversion to ada map yet')



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ada.base.physical_objects import BackendGeom
from ada.base.units import Units
from ada.concepts.bounding_box import BoundingBox
from ada.concepts.curves import CurvePoly
from ada.concepts.points import Node
from ada.concepts.transforms import Placement
from ada.config import Settings
from ada.materials import Material
from ada.materials.metals import CarbonSteel

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore


class Plate(BackendGeom):
    """
    A plate object. The plate element covers all plate elements. Contains a dictionary with each point of the plate
    described by an id (index) and a Node object.

    :param name: Name of plate
    :param nodes: List of coordinates that make up the plate. Points can be Node, tuple or list
    :param t: Thickness of plate
    :param mat: Material. Can be either Material object or built-in materials ('S420' or 'S355')
    :param placement: Explicitly define origin of plate. If not set
    """

    def __init__(
        self,
        name,
        nodes,
        t,
        mat="S420",
        use3dnodes=False,
        placement=Placement(),
        pl_id=None,
        offset=None,
        colour=None,
        parent=None,
        ifc_geom=None,
        opacity=1.0,
        metadata=None,
        tol=None,
        units=Units.M,
        guid=None,
        ifc_store: IfcStore = None,
    ):
        super().__init__(
            name,
            guid=guid,
            metadata=metadata,
            units=units,
            placement=placement,
            ifc_store=ifc_store,
            colour=colour,
            opacity=opacity,
        )

        points2d = None
        points3d = None

        if use3dnodes is True:
            points3d = nodes
        else:
            points2d = nodes

        self._pl_id = pl_id
        self._material = mat if isinstance(mat, Material) else Material(mat, mat_model=CarbonSteel(mat), parent=parent)
        self._material.refs.append(self)
        self._t = t

        if tol is None:
            tol = Units.get_general_point_tol(units)

        self._poly = CurvePoly(
            points3d=points3d,
            points2d=points2d,
            normal=self.placement.zdir,
            origin=self.placement.origin,
            xdir=self.placement.xdir,
            tol=tol,
            parent=self,
        )

        self._offset = offset
        self._parent = parent
        self._ifc_geom = ifc_geom
        self._bbox = None

    @property
    def id(self):
        return self._pl_id

    @id.setter
    def id(self, value):
        self._pl_id = value

    @property
    def offset(self):
        return self._offset

    @property
    def t(self) -> float:
        """Plate thickness"""
        return self._t

    @property
    def material(self) -> Material:
        return self._material

    @material.setter
    def material(self, value: Material):
        self._material = value

    @property
    def n(self) -> np.ndarray:
        """Normal vector"""
        return self.poly.normal

    @property
    def nodes(self) -> list[Node]:
        return self.poly.nodes

    @property
    def poly(self) -> CurvePoly:
        return self._poly

    def bbox(self) -> BoundingBox:
        """Bounding Box of plate"""
        if self._bbox is None:
            self._bbox = BoundingBox(self)

        return self._bbox

    def line(self):
        return self._poly.wire

    def shell(self):
        from ada.occ.utils import apply_penetrations

        geom = apply_penetrations(self.poly.face, self.penetrations)

        return geom

    def solid(self):
        from ada.occ.utils import apply_penetrations

        geom = apply_penetrations(self._poly.make_extruded_solid(self.t), self.penetrations)

        return geom

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if self._units != value:
            scale_factor = Units.get_scale_factor(self._units, value)
            tol = Settings.mmtol if value == "mm" else Settings.mtol
            self._t *= scale_factor
            self.poly.scale(scale_factor, tol)
            for pen in self.penetrations:
                pen.units = value
            self.material.units = value
            self._units = value

    def __repr__(self):
        return f"Plate({self.name}, t:{self.t}, {self.material})"


# https://standards.buildingsmart.org/IFC/RELEASE/IFC4_3/lexical/IfcBSplineSurfaceWithKnots.htm



from __future__ import annotations

from typing import TYPE_CHECKING

import ifcopenshell.geom
import numpy as np

from ada.ifc.utils import create_guid
from ada.visualize.concept import ObjectMesh, PartMesh, VisMesh

if TYPE_CHECKING:
    from ada import Part


def part_to_vis_mesh2(part: Part, auto_sync_ifc_store=True, cpus: int = None) -> VisMesh:
    ifc_store = part.get_assembly().ifc_store
    if auto_sync_ifc_store:
        ifc_store.sync()

    settings = ifcopenshell.geom.settings()
    settings.set(settings.USE_PYTHON_OPENCASCADE, False)
    settings.set(settings.SEW_SHELLS, False)
    settings.set(settings.WELD_VERTICES, True)
    settings.set(settings.INCLUDE_CURVES, False)
    settings.set(settings.USE_WORLD_COORDS, True)
    settings.set(settings.VALIDATE_QUANTITIES, False)

    id_map = dict()

    res = list(ifc_store.assembly.get_all_physical_objects())

    if len(res) > 0:
        iterator = ifc_store.get_ifc_geom_iterator(settings, cpus=cpus)
        iterator.initialize()
        while True:
            shape = iterator.get()
            if shape:
                obj_mesh = product_to_obj_mesh(shape)
                id_map[shape.guid] = obj_mesh

            if not iterator.next():
                break

    pm = PartMesh(name=part.name, id_map=id_map)
    meta = {
        p.guid: (p.name, p.parent.name if p.parent is not None else "*")
        for p in part.get_all_subparts(include_self=True)
    }
    parts_d = {p.guid: (p.name, p.parent.guid) for p in part.get_all_physical_objects()}
    meta.update(parts_d)

    for p in part.get_all_subparts(include_self=True):
        if p.fem.is_empty() is True:
            continue

        mesh = p.fem.to_mesh()
        coords = mesh.nodes.coords
        edges, faces = mesh.get_edges_and_faces_from_mesh()
        guid = create_guid()
        name = p.fem.name
        meta.update({guid: (name, p.guid)})

        id_map.update({guid: ObjectMesh(guid, faces, coords, edges=edges)})

    return VisMesh(part.name, world=[pm], meta=meta)


def product_to_obj_mesh(shape: ifcopenshell.ifcopenshell_wrapper.TriangulationElement) -> ObjectMesh:
    geometry = shape.geometry
    vertices = np.array(geometry.verts, dtype="float32").reshape(int(len(geometry.verts) / 3), 3)
    faces = np.array(geometry.faces, dtype=int)
    normals = np.array(geometry.normals) if len(geometry.normals) != 0 else None

    if normals is not None and len(normals) > 0:
        normals = normals.astype(dtype="float32").reshape(int(len(normals) / 3), 3)

    mats = geometry.materials
    if len(mats) == 0:
        # colour = [1.0, 0.0, 0.0, 1.0]
        colour = None
    else:
        mat0 = mats[0]
        opacity = 1.0 - mat0.transparency
        if mat0.diffuse == (0.0, 0.0, 0.0):
            colour = None
        else:
            colour = [*mat0.diffuse, opacity]

    return ObjectMesh(shape.guid, faces, vertices, normals, color=colour)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

import ifcopenshell
import ifcopenshell.geom

from ada.base.changes import ChangeAction
from ada.ifc.read.reader_utils import get_ifc_body
from ada.ifc.utils import add_negative_extrusion, create_guid, write_elem_property_sets
from ada.ifc.write.write_beams import write_ifc_beam
from ada.ifc.write.write_fasteners import write_ifc_fastener
from ada.ifc.write.write_instances import write_mapped_instance
from ada.ifc.write.write_material import write_ifc_mat
from ada.ifc.write.write_openings import generate_ifc_opening
from ada.ifc.write.write_pipe import write_ifc_pipe
from ada.ifc.write.write_plates import write_ifc_plate
from ada.ifc.write.write_sections import export_beam_section_profile_def
from ada.ifc.write.write_shapes import write_ifc_shape
from ada.ifc.write.write_spatial_elements import (
    write_ifc_part,
    write_ifc_spatial_hierarchy,
)
from ada.ifc.write.write_wall import write_ifc_wall

if TYPE_CHECKING:
    from ada import Beam, Material, Part, Pipe, Plate, Section, Shape, Wall
    from ada.ifc.store import IfcStore


def is_added(x):
    return x.change_type == ChangeAction.ADDED


def is_modified(x):
    return x.change_type == ChangeAction.MODIFIED


def is_deleted(x):
    return x.change_type == ChangeAction.DELETED


@dataclass
class IfcWriter:
    ifc_store: IfcStore
    callback: Callable[[int, int], None] = None

    def sync_spatial_hierarchy(self, include_fem=False) -> int:
        if len(list(self.ifc_store.f.by_type("IfcSite"))) == 0:
            write_ifc_spatial_hierarchy(self.ifc_store)

        num_new_spatial_objects = 0
        for part in filter(is_added, self.ifc_store.assembly.get_all_parts_in_assembly()):
            self.add_part(part, include_fem=include_fem)
            part.change_type = ChangeAction.NOCHANGE
            num_new_spatial_objects += 1
        return num_new_spatial_objects

    def sync_added_physical_objects(self) -> int:
        a = self.ifc_store.assembly
        mat_map = {mat.guid: mat for mat in a.get_all_materials()}
        rel_mats_map = {
            m.GlobalId: m
            for m in filter(
                lambda x: x.RelatingMaterial.is_a("IfcMaterial"), self.ifc_store.f.by_type("IfcRelAssociatesMaterial")
            )
        }

        num_new_objects = len(list(a.get_all_physical_objects()))
        contained_in_spatial = {x.guid: [] for x in a.get_all_parts_in_assembly(include_self=True)}

        for i, to_be_added in enumerate(filter(is_added, a.get_all_physical_objects()), start=1):
            self.eval_validity(to_be_added, mat_map, rel_mats_map)

            ifc_elem = self.add(to_be_added)
            self.create_ifc_openings(to_be_added, ifc_elem)

            write_elem_property_sets(to_be_added.metadata, ifc_elem, self.ifc_store.f, self.ifc_store.owner_history)

            contained_in_spatial[to_be_added.parent.guid].append(ifc_elem)
            to_be_added.change_type = ChangeAction.NOCHANGE
            if self.callback is not None:
                self.callback(i, num_new_objects)

        for spatial_elem_guid, relating_elements in contained_in_spatial.items():
            if len(relating_elements) == 0:
                continue
            self.add_related_elements_to_spatial_container(relating_elements, spatial_elem_guid)

        return num_new_objects

    def sync_modified_physical_objects(self) -> int:
        num_mod = 0
        for to_be_modified in filter(is_modified, self.ifc_store.assembly.get_all_physical_objects()):
            self.create_ifc_openings(to_be_modified)
            to_be_modified.change_type = ChangeAction.NOCHANGE
            num_mod += 1
        return num_mod

    def sync_added_welds(self):
        for weld in self.ifc_store.assembly.welds:
            ifc_weld = write_ifc_fastener(weld)
            self.add_related_elements_to_spatial_container([ifc_weld], weld.parent.guid)
            if weld.groove is None:
                continue

            for mem in weld.members:
                rel_ifc_mem = self.ifc_store.f.by_guid(mem.guid)
                ifc_opening = generate_ifc_opening(weld.groove)
                self.ifc_store.f.create_entity(
                    "IfcRelVoidsElement",
                    GlobalId=create_guid(),
                    OwnerHistory=self.ifc_store.owner_history,
                    Name=None,
                    Description=None,
                    RelatingBuildingElement=rel_ifc_mem,
                    RelatedOpeningElement=ifc_opening,
                )

    def sync_deleted_physical_objects(self) -> int:
        num_mod = 0
        for to_be_modified in filter(is_deleted, self.ifc_store.assembly.get_all_physical_objects()):
            self.create_ifc_openings(to_be_modified)
            to_be_modified.change_type = ChangeAction.NOCHANGE
            num_mod += 1
        return num_mod

    def sync_presentation_layers(self) -> int:
        from ada import Part, Penetration, Pipe

        num_added = 0
        f = self.ifc_store.f

        def append_bodies(ifc_ref: str | ifcopenshell.entity_instance, assigned_items_):
            if isinstance(ifc_ref, str):
                ifc_obj_ = f.by_guid(ifc_ref)
            else:
                ifc_obj_ = ifc_ref

            ifc_body_ = get_ifc_body(ifc_obj_, allow_multiple=True)
            if isinstance(ifc_body_, list):
                for body_ in ifc_body_:
                    assigned_items_.append(body_)
            else:
                assigned_items_.append(ifc_body_)

        for layer in self.ifc_store.assembly.presentation_layers.layers.values():
            assigned_items = []
            for member in layer.members:
                if isinstance(member, Pipe):
                    for seg in member.segments:
                        append_bodies(seg.guid, assigned_items)

                elif isinstance(member, Part):
                    continue
                elif isinstance(member, Penetration):
                    for ifc_opening in f.by_type("IfcOpeningElement"):
                        if ifc_opening.Name != member.name:
                            continue
                        append_bodies(ifc_opening, assigned_items)

                else:
                    append_bodies(member.guid, assigned_items)

            exist_layer = None
            for ifc_layer in f.by_type("IfcPresentationLayerAssignment"):
                if ifc_layer.Name == layer.name:
                    exist_layer = ifc_layer
                    break
            if len(assigned_items) == 0:
                continue

            if exist_layer is None:
                f.create_entity(
                    "IfcPresentationLayerAssignment",
                    Name=layer.name,
                    Description=layer.description,
                    AssignedItems=assigned_items,
                    Identifier=layer.identifier,
                    # LayerOn=True,
                    # LayerFrozen=False,
                    # LayerBlocked=False,
                    # LayerStyles=[presentation_style],
                )
            else:
                updated_assigned_items = list(exist_layer.AssignedItems)
                for ai in assigned_items:
                    if ai not in updated_assigned_items:
                        updated_assigned_items.append(ai)
                exist_layer.AssignedItems = updated_assigned_items

            layer.change_type = ChangeAction.NOCHANGE

        return num_added

    def sync_mapped_instances(self):
        for part in self.ifc_store.assembly.get_all_parts_in_assembly(include_self=True):
            for instance in part.instances.values():
                write_mapped_instance(instance, self.ifc_store.f)

    def sync_sections(self):
        for sec in self.ifc_store.assembly.get_all_sections():
            if sec.change_type != ChangeAction.ADDED:
                continue

            self.create_ifc_profile_def(sec)
            self.create_ifc_beam_type(sec)
            sec.change_type = ChangeAction.NOCHANGE

    def sync_materials(self):
        all_mats = self.ifc_store.assembly.get_all_materials()
        skipped_mats = []
        for mat in all_mats:
            if mat.change_type != ChangeAction.ADDED:
                skipped_mats.append(mat)
                continue
            self.create_ifc_material(mat)
            mat.change_type = ChangeAction.NOCHANGE

        skip_mats = set([m.guid for m in skipped_mats])
        mat_map = {mat.guid for mat in self.ifc_store.assembly.get_all_materials()} - skip_mats

        rel_mats_map = {
            m.GlobalId
            for m in filter(
                lambda x: x.RelatingMaterial.is_a("IfcMaterial"), self.ifc_store.f.by_type("IfcRelAssociatesMaterial")
            )
        }
        if len(mat_map - rel_mats_map) != 0:
            raise ValueError("Syncing of Materials failed")

    def create_ifc_openings(self, obj: Beam | Plate | Pipe | Shape | Wall, ifc_obj=None):
        from ada import Part, Wall
        from ada.core.constants import O, X, Z

        f = self.ifc_store.f

        if ifc_obj is None:
            ifc_obj = f.by_guid(obj.guid)

        if isinstance(obj, Wall):
            if len(obj.inserts) > 0:
                for i, insert in enumerate(obj.inserts):
                    add_negative_extrusion(f, O, Z, X, insert.height, obj.openings_extrusions[i], ifc_obj)
                    if issubclass(type(insert), Part) is False:
                        raise ValueError(f'Unrecognized type "{type(insert)}"')

        else:
            if len(obj.penetrations) > 0:
                for pen in obj.penetrations:
                    ifc_opening = generate_ifc_opening(pen)
                    f.create_entity(
                        "IfcRelVoidsElement",
                        GlobalId=create_guid(),
                        OwnerHistory=self.ifc_store.owner_history,
                        Name=None,
                        Description=None,
                        RelatingBuildingElement=ifc_obj,
                        RelatedOpeningElement=ifc_opening,
                    )

    def eval_validity(self, to_be_added, mat_map, rel_mats_map):
        from ada import Pipe, Shape, Wall

        if isinstance(to_be_added, Wall) is False and issubclass(type(to_be_added), Shape) is False:
            if to_be_added.material.guid not in mat_map.keys():
                raise ValueError(f"Object {to_be_added.material} is not among synced materials {mat_map}")
            if to_be_added.material.guid not in rel_mats_map.keys():
                raise ValueError(f"Object {to_be_added.material} is not among synced materials {rel_mats_map}")

        elif isinstance(to_be_added, Pipe):
            for seg in to_be_added.segments:
                if seg.material.guid not in mat_map.keys():
                    raise ValueError(f"Object {to_be_added.material} is not among synced materials {mat_map}")
                if seg.material.guid not in rel_mats_map.keys():
                    raise ValueError(f"Object {to_be_added.material} is not among synced materials {rel_mats_map}")

    def add_related_elements_to_spatial_container(self, elements: list[ifcopenshell.entity_instance], guid: str):
        parent_ifc_elem = self.ifc_store.get_by_guid(guid)

        existing_spatial = None
        for existing_rel in self.ifc_store.f.by_type("IfcRelContainedInSpatialStructure"):
            if parent_ifc_elem == existing_rel.RelatingStructure:
                existing_spatial = existing_rel
                break

        if existing_spatial is not None:
            existing_spatial.OwnerHistory = self.ifc_store.owner_history
            existing_spatial.RelatedElements = list(existing_spatial.RelatedElements) + elements
        else:
            self.ifc_store.f.create_entity(
                "IfcRelContainedInSpatialStructure",
                GlobalId=create_guid(),
                OwnerHistory=self.ifc_store.owner_history,
                Name="Physical model",
                Description=None,
                RelatedElements=elements,
                RelatingStructure=parent_ifc_elem,
            )

    def associate_elem_with_material(self, material: Material, ifc_elem: ifcopenshell.entity_instance):
        try:
            rel_mat = self.ifc_store.f.by_guid(material.guid)
        except RuntimeError as e:
            raise RuntimeError(e)
        related_objects = [*rel_mat.RelatedObjects, ifc_elem]
        rel_mat.RelatedObjects = related_objects
        return rel_mat

    def associate_elem_with_profiledef(self, section: Section, ifc_elem: ifcopenshell.entity_instance):
        """This is only for IFC 4.3++"""
        rel_profile_def = self.ifc_store.f.by_guid(section.guid)
        related_objects = [*rel_profile_def.RelatedObjects, ifc_elem]
        rel_profile_def.RelatedObjects = related_objects
        return rel_profile_def

    def add(self, obj: Beam | Plate | Pipe | Shape | Wall) -> ifcopenshell.entity_instance:
        from ada import Beam, Pipe, Plate, Shape, Wall

        if isinstance(obj, Beam):
            return write_ifc_beam(self.ifc_store, obj)
        elif isinstance(obj, Plate):
            return write_ifc_plate(obj)
        elif isinstance(obj, Pipe):
            return write_ifc_pipe(obj)
        elif issubclass(type(obj), Shape):
            return write_ifc_shape(obj)
        elif isinstance(obj, Wall):
            return write_ifc_wall(obj)
        else:
            raise NotImplementedError()

    def add_part(self, part: Part, include_fem):
        return write_ifc_part(self.ifc_store, part, include_fem=include_fem)

    def create_ifc_profile_def(self, section: Section):
        export_beam_section_profile_def(section)

    def create_ifc_beam_type(self, section: Section):
        self.ifc_store.f.create_entity(
            "IfcBeamType",
            GlobalId=section.guid,
            OwnerHistory=self.ifc_store.owner_history,
            Name=section.name,
            Description=section.sec_str,
            PredefinedType="BEAM",
        )

    def create_ifc_material(self, material: Material):
        ifc_mat = write_ifc_mat(material)
        self.create_rel_associates_material(material.guid, ifc_mat)
        return ifc_mat

    def create_rel_associates_material(self, guid: str, relating_mat: ifcopenshell.entity_instance, related_objs=None):
        return self.ifc_store.f.create_entity(
            "IfcRelAssociatesMaterial",
            GlobalId=guid,
            OwnerHistory=self.ifc_store.owner_history,
            Name=relating_mat.Name if hasattr(relating_mat, "Name") else None,
            Description=f"Objects related to {relating_mat.Name}" if hasattr(relating_mat, "Description") else None,
            RelatedObjects=[] if related_objs is None else related_objs,
            RelatingMaterial=relating_mat,
        )



from ada.base.units import Units
from ada.config import get_logger
from ada.sections import Section

logger = get_logger()


def import_section_from_ifc(profile_def, units=Units.M) -> Section:
    """Takes any subclass of ProfileDef"""
    from ada.sections.utils import interpret_section_str

    if profile_def.is_a("IfcIShapeProfileDef"):
        sec = Section(
            name=profile_def.ProfileName,
            sec_type=Section.TYPES.IPROFILE,
            h=profile_def.OverallDepth,
            w_top=profile_def.OverallWidth,
            w_btn=profile_def.OverallWidth,
            t_w=profile_def.WebThickness,
            t_ftop=profile_def.FlangeThickness,
            t_fbtn=profile_def.FlangeThickness,
            units=units,
            sec_str=profile_def.ProfileName,
        )
    elif profile_def.is_a("IfcTShapeProfileDef"):
        sec = Section(
            name=profile_def.ProfileName,
            sec_type=Section.TYPES.TPROFILE,
            h=profile_def.Depth,
            w_top=profile_def.FlangeWidth,
            t_w=profile_def.WebThickness,
            t_ftop=profile_def.FlangeThickness,
            units=units,
        )
    elif profile_def.is_a("IfcCircleHollowProfileDef"):
        sec = Section(
            name=profile_def.ProfileName,
            sec_type="TUB",
            r=profile_def.Radius,
            wt=profile_def.WallThickness,
            units=units,
        )
    elif profile_def.is_a("IfcUShapeProfileDef"):
        raise NotImplementedError(f'IFC section type "{profile_def}" is not yet implemented')
        # sec = Section(ifc_elem.ProfileName)
    else:
        try:
            logger.info(f'No Native support for Ifc beam "{profile_def=}"')
            sec, tap = interpret_section_str(profile_def.ProfileName)
        except ValueError as e:
            logger.warning(f'Unable to process section "{profile_def.ProfileName}" -> error: "{e}" ')
            sec = None
        if sec is None:
            raise NotImplementedError(f'IFC section type "{profile_def}" is not yet implemented')

    return sec



from __future__ import annotations

import reprlib
from bisect import bisect_left, bisect_right
from itertools import chain
from operator import attrgetter
from typing import TYPE_CHECKING, Dict, Iterable, List, Union

import numpy as np

from ada.base.units import Units
from ada.concepts.exceptions import DuplicateNodes
from ada.concepts.points import Node, replace_node
from ada.concepts.stru_beams import Beam
from ada.concepts.stru_plates import Plate
from ada.concepts.transforms import Rotation
from ada.config import Settings, get_logger
from ada.core.utils import Counter, roundoff
from ada.core.vector_utils import (
    is_null_vector,
    is_parallel,
    points_in_cylinder,
    unit_vector,
    vector_length,
)
from ada.materials import Material

if TYPE_CHECKING:
    from ada import FEM, Assembly, Part
    from ada.concepts.connections import JointBase
    from ada.fem.results.common import FemNodes
    from ada.sections import Section

logger = get_logger()

__all__ = [
    "Nodes",
    "Beams",
    "Plates",
    "Connections",
    "Materials",
    "Sections",
]


class BaseCollections:
    """The Base class for all collections"""

    def __init__(self, parent: Part):
        self._parent = parent

    @property
    def parent(self) -> Part:
        return self._parent


class Beams(BaseCollections):
    """A collections of Beam objects"""

    def __init__(self, beams: Iterable[Beam] = None, parent=None):
        super().__init__(parent)
        beams = [] if beams is None else beams
        self._beams = sorted(beams, key=attrgetter("name"))
        self._nmap = {n.name: n for n in self._beams}
        self._dmap = {n.guid: n for n in self._beams}
        self._connected_beams_map = None

    def __contains__(self, item: Beam):
        return item.guid in self._dmap.keys()

    def __len__(self):
        return len(self._beams)

    def __iter__(self) -> Iterable[Beam]:
        return iter(self._beams)

    def __getitem__(self, index):
        result = self._beams[index]
        return Beams(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Beams):
            return NotImplemented
        return self._beams == other._beams

    def __ne__(self, other):
        if not isinstance(other, Beams):
            return NotImplemented
        return self._beams != other._beams

    def __add__(self, other):
        return Beams(chain(self, other))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Beams({rpr.repr(self._beams) if self._beams else ''})"

    def merge_connected_beams_by_properties(self) -> None:
        def append_connected_beams(connected_beams: Iterable[Beam]) -> None:
            for c_beam in connected_beams:
                if c_beam not in to_be_merged:
                    to_be_merged.append(c_beam)
                    append_connected_beams(self.connected_beams_map[c_beam])

        self.set_connected_beams_map()
        merged_beams: list[Beam] = list()

        for beam in self._beams.copy():
            if beam not in merged_beams:
                to_be_merged: list[Beam] = [beam]
                append_connected_beams(self.connected_beams_map[beam])
                merged_beams.extend(to_be_merged)
                self.merge_beams(to_be_merged)

        self.set_connected_beams_map()

    def merge_beams(self, beam_segments: Iterable[Beam]) -> Beam:
        """Merge all beam segments into the first entry in beam_segments by changing the beam nodes."""

        def get_end_nodes() -> list[Node]:
            end_beams = filter(lambda x: len(self.connected_beams_map.get(x, list())) == 1, beam_segments)

            end_nds: list[Node] = list()

            for beam in end_beams:
                (node_without_connected_beam,) = self.connected_beams_map[beam]
                end_nds.append(beam.n1 if node_without_connected_beam in beam.n2.refs else beam.n2)
            return end_nds

        def modify_beam(bm: Beam, new_nodes) -> Beam:
            n1, n2 = new_nodes

            n1_2_n2_vector = unit_vector(n2.p - n1.p)
            beam_vector = bm.xvec.round(decimals=Settings.precision)

            if is_parallel(n1_2_n2_vector, bm.xvec) and not is_null_vector(n1_2_n2_vector, bm.xvec):
                n1, n2 = n2, n1
            elif not is_parallel(n1_2_n2_vector, bm.xvec):
                raise ValueError(f"Unit vector error. Beam.xvec: {beam_vector}, nodes unit_vec: {-1 * n1_2_n2_vector}")

            bm.n1, bm.n2 = n1, n2
            return bm

        if len(list(beam_segments)) > 1:
            end_nodes = get_end_nodes()
            modified_beam = modify_beam(beam_segments[0], end_nodes)

            for old_beam in beam_segments[1:]:
                self.remove(old_beam)

            return modified_beam

    def set_connected_beams_map(self) -> None:
        self._connected_beams_map = {beam: beam.get_beam_extensions() for beam in self._beams}

    @property
    def connected_beams_map(self) -> dict[Beam, Iterable[Beam]]:
        return self._connected_beams_map

    def get_beams_at_point(self, point: Union[Node, np.ndarray]) -> list[Beam]:
        return list(filter(lambda x: x.is_point_on_beam(point), self._beams))

    def index(self, item: Beam) -> int:
        index = bisect_left(self._beams, item)
        if (index != len(self._beams)) and (self._beams[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item) -> int:
        return int(item in self)

    def from_name(self, name: str) -> Beam:
        """Get beam from its name"""
        return self._nmap.get(name)

    def add(self, beam: Beam) -> Beam:
        from .exceptions import NameIsNoneError

        if beam.name is None:
            raise NameIsNoneError("Name is not allowed to be None.")

        if beam.name in self._dmap.keys():
            logger.warning(f'Beam with name "{beam.name}" already exists. Will not add')
            return self._dmap[beam.name]

        self._dmap[beam.guid] = beam
        self._nmap[beam.name] = beam
        self._beams.append(beam)
        beam.add_beam_to_node_refs()
        return beam

    def remove(self, beam: Beam) -> None:
        beam.remove_beam_from_node_refs()
        i = self._beams.index(beam)
        self._beams.pop(i)
        self._dmap = {n.guid: n for n in self._beams}
        self._nmap = {n.name: n for n in self._beams}

    def get_beams_within_volume(self, vol_, margins=Settings.point_tol) -> Iterable[Beam]:
        """
        :param vol_: List or tuple of tuples [(xmin, xmax), (ymin, ymax), (zmin, zmax)]
        :param margins: Add margins to the volume box (equal in all directions). Input is in meters. Can be negative.
        :return: List of beam ids
        """
        from bisect import bisect_left, bisect_right

        if margins is not None:
            vol_new = []
            for p in vol_:
                vol_new.append((roundoff(p[0] - margins), roundoff(p[1] + margins)))
        else:
            vol_new = vol_
        vol = vol_new

        def sort_beams(bms):
            xkeys = [key[1] for key in bms]
            xmin = bisect_left(xkeys, vol[0][0])
            xmax = bisect_right(xkeys, vol[0][1])

            within_x_list = sorted(bms[xmin:xmax], key=lambda elem: elem[2])

            ykeys = [key[2] for key in within_x_list]
            ymin = bisect_left(ykeys, vol[1][0])
            ymax = bisect_right(ykeys, vol[1][1])

            within_y_list = sorted(within_x_list[ymin:ymax], key=lambda elem: elem[3])

            zkeys = [key[3] for key in within_y_list]
            zmin = bisect_left(zkeys, vol[2][0])
            zmax = bisect_right(zkeys, vol[2][1])

            within_vol_list = within_y_list[zmin:zmax]
            return [bm[0] for bm in within_vol_list]

        bm_list1 = [(bm.name, bm.n1.x, bm.n1.y, bm.n1.z) for bm in sorted(self._beams, key=lambda bm: bm.n1.x)]
        bm_list2 = [(bm.name, bm.n2.x, bm.n2.y, bm.n2.z) for bm in sorted(self._beams, key=lambda bm: bm.n2.x)]

        return set([self.from_name(bm_id) for bms_ in (bm_list1, bm_list2) for bm_id in sort_beams(bms_)])

    @property
    def dmap(self) -> dict[str, Beam]:
        return self._dmap

    @property
    def nmap(self) -> dict[str, Beam]:
        return self._nmap


class Plates(BaseCollections):
    """Plate object collection"""

    def __init__(self, plates: Iterable[Plate] = None, parent: Part = None):
        plates = [] if plates is None else plates
        super().__init__(parent)
        self._plates = sorted(plates, key=attrgetter("name"))
        self._idmap = {n.guid: n for n in self._plates}
        self._nmap = {n.name: n for n in self._plates}

    def __contains__(self, item: Plate):
        return item.guid in self._idmap.keys()

    def __len__(self):
        return len(self._plates)

    def __iter__(self) -> Iterable[Plate]:
        return iter(self._plates)

    def __getitem__(self, index):
        result = self._plates[index]
        return Materials(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Plates):
            return NotImplemented
        return self._plates == other._plates

    def __ne__(self, other):
        if not isinstance(other, Plates):
            return NotImplemented
        return self._plates != other._plates

    def __add__(self, other: Plates) -> Plates:
        return Plates(chain(self, other))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Plates({rpr.repr(self._plates) if self._plates else ''})"

    def index(self, plate: Plate):
        index = bisect_left(self._plates, plate)
        if (index != len(self._plates)) and (self._plates[index] == plate):
            return index
        raise ValueError(f"{repr(plate)} not found")

    def count(self, item: Plate):
        return int(item in self)

    def remove(self, plate: Plate) -> None:
        i = self._plates.index(plate)
        self._plates.pop(i)
        self._idmap = {n.guid: n for n in self._plates}
        self._nmap = {n.name: n for n in self._plates}

    def from_name(self, name: str) -> Plate:
        return self._nmap.get(name, None)

    @property
    def dmap(self) -> dict[str, Plate]:
        return self._idmap

    @property
    def nmap(self) -> dict[str, Plate]:
        return self._nmap

    def add(self, plate: Plate) -> Plate:
        if plate.name is None:
            raise Exception("Name is not allowed to be None.")

        if plate.name in self._nmap.keys():
            return self._nmap[plate.name]
        mat = self._parent.materials.add(plate.material)
        if mat is not None:
            plate.material = mat

        self._plates.append(plate)
        self._nmap[plate.name] = plate
        self._idmap[plate.guid] = plate
        return plate


class Connections(BaseCollections):
    _counter = Counter(1, "C")

    def __init__(self, connections: Iterable[JointBase] = None, parent=None):
        connections = [] if connections is None else connections
        super().__init__(parent)
        self._connections = connections
        self._initialize_connection_data()

    def _initialize_connection_data(self):
        self._dmap = {j.name: j for j in self._connections}
        self._joint_centre_nodes = Nodes([c.centre for c in self._connections])
        self._nmap = {self._joint_centre_nodes.index(c.centre): c for c in self._connections}

    @property
    def connections(self) -> List[JointBase]:
        return self._connections

    @connections.setter
    def connections(self, value: List[JointBase]):
        self._connections = value
        self._initialize_connection_data()

    @property
    def joint_centre_nodes(self):
        return self._joint_centre_nodes

    def __contains__(self, item):
        return item.id in self._dmap.keys()

    def __len__(self):
        return len(self._connections)

    def __iter__(self) -> Iterable[JointBase]:
        return iter(self._connections)

    def __getitem__(self, index):
        result = self._connections[index]
        return Connections(result) if isinstance(index, slice) else result

    def __eq__(self, other: Connections):
        if not isinstance(other, Connections):
            return NotImplemented
        return self._connections == other._connections

    def __ne__(self, other: Connections):
        if not isinstance(other, Connections):
            return NotImplemented
        return self._connections != other._connections

    def __add__(self, other: Connections):
        return Connections(chain(self._connections, other._connections))

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Connections({rpr.repr(self._connections) if self._connections else ''})"

    def get_from_name(self, name: str):
        result = self._dmap.get(name, None)
        if result is None:
            logger.error(f'No Joint with the name "{name}" found within this connection object')
        return result

    def add(self, joint: JointBase, point_tol=Settings.point_tol):
        if joint.name is None:
            raise Exception("Name is not allowed to be None.")

        if joint.name in self._dmap.keys():
            raise ValueError("Joint Exists with same name")

        new_node = Node(joint.centre)
        node = self._joint_centre_nodes.add(new_node, point_tol=point_tol)
        if node != new_node:
            return self._nmap[node]
        else:
            self._nmap[node] = joint
        joint.parent = self
        self._dmap[joint.name] = joint
        self._connections.append(joint)

    def remove(self, joint: JointBase):
        if joint.name in self._dmap.keys():
            self._dmap.pop(joint.name)
        if joint in self._connections:
            self._connections.pop(self._connections.index(joint))
        if joint.centre in self._nmap.keys():
            self._nmap.pop(joint.centre)

    def find(self, out_of_plane_tol=0.1, joint_func=None, point_tol=Settings.point_tol):
        """
        Find all connections between beams in all parts using a simple clash check.

        :param out_of_plane_tol:
        :param joint_func: Pass a function for mapping the generic Connection classes to a specific reinforced Joints
        :param point_tol:
        """
        from ada.concepts.connections import JointBase
        from ada.core.clash_check import are_beams_connected

        ass = self._parent.get_assembly()
        bm_res = ass.beam_clash_check()

        nodes = Nodes()
        nmap = dict()

        for bm1_, beams_ in bm_res:
            are_beams_connected(bm1_, beams_, out_of_plane_tol, point_tol, nodes, nmap)

        for node, mem in nmap.items():
            if joint_func is not None:
                joint = joint_func(next(self._counter), mem, node.p, parent=self)
                if joint is None:
                    continue
            else:
                joint = JointBase(next(self._counter), mem, node.p, parent=self)

            self.add(joint, point_tol=point_tol)

        print(f"Connection search finished. Found a total of {len(self._connections)} connections")


class NumericMapped(BaseCollections):
    def __init__(self, parent):
        super(NumericMapped, self).__init__(parent=parent)
        self._name_map = dict()
        self._id_map = dict()

    def recreate_name_and_id_maps(self, collection):
        self._name_map = {n.name: n for n in collection}
        self._id_map = {n.id: n for n in collection}

    @property
    def max_id(self):
        if len(self._id_map.keys()) == 0:
            return 0
        return max(self._id_map.keys())


class Materials(NumericMapped):
    """Collection of materials"""

    def __init__(self, materials: Iterable[Material] = None, parent: Union[Part, Assembly] = None, units=Units.M):
        super().__init__(parent)
        self.materials = sorted(materials, key=attrgetter("name")) if materials is not None else []
        self.recreate_name_and_id_maps(self.materials)
        self._units = units

    def __contains__(self, item: Material):
        return item.name in self._name_map.keys()

    def __len__(self) -> int:
        return len(self.materials)

    def __iter__(self) -> Iterable[Material]:
        return iter(self.materials)

    def __getitem__(self, index):
        result = self.materials[index]
        return Materials(result) if isinstance(index, slice) else result

    def __eq__(self, other: Materials):
        if not isinstance(other, Materials):
            return NotImplemented
        return self.materials == other.materials

    def __ne__(self, other: Materials):
        if not isinstance(other, Materials):
            return NotImplemented
        return self.materials != other.materials

    def __add__(self, other: Materials):
        if self.parent is None:
            raise ValueError("Parent cannot be zero")
        for mat in other:
            mat.parent = self.parent
        other.renumber_id(self.max_id + 1)
        return Materials(chain(self, other), parent=self.parent)

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Materials({rpr.repr(self.materials) if self.materials else ''})"

    def merge_materials_by_properties(self):
        models = []

        final_mats = []
        for i, mat in enumerate(self.materials):
            if mat.model.unique_props() not in models:
                models.append(mat.model.unique_props())
                final_mats.append(mat)
            else:
                index = models.index(mat.model.unique_props())
                replacement_mat = final_mats[index]
                for ref in mat.refs:
                    ref.material = replacement_mat

        self.materials = final_mats
        self.recreate_name_and_id_maps(self.materials)

    def index(self, item: Material):
        return self.materials.index(item)

    def count(self, item: Material):
        return int(item in self)

    def get_by_name(self, name: str) -> Material:
        if name not in self._name_map.keys():
            raise ValueError(f'The material name "{name}" is not found')
        else:
            return self._name_map[name]

    def get_by_id(self, mat_id: int) -> Material:
        if mat_id not in self._id_map.keys():
            raise ValueError(f'The material id "{mat_id}" is not found')
        else:
            return self._id_map[mat_id]

    def renumber_id(self, start_id=1):
        cnt = Counter(start=start_id)
        for mat_id in sorted(self.id_map.keys()):
            mat = self.get_by_id(mat_id)
            mat.id = next(cnt)
        self.recreate_name_and_id_maps(self.materials)

    @property
    def name_map(self) -> Dict[str, Material]:
        return self._name_map

    @property
    def id_map(self) -> Dict[int, Material]:
        return self._id_map

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            for m in self.materials:
                m.units = value
            self._units = value

    def add(self, material) -> Material:
        if material in self:
            existing_mat = self._name_map[material.name]
            for elem in material.refs:
                if elem not in existing_mat.refs:
                    existing_mat.refs.append(elem)
            return existing_mat

        if material.id is None or material.id in self._id_map.keys():
            material.id = len(self.materials) + 1

        self._id_map[material.id] = material
        self._name_map[material.name] = material
        self.materials.append(material)

        return material


class Sections(NumericMapped):
    def __init__(self, sections: Iterable[Section] = None, parent: Part | Assembly = None, units=Units.M):
        sec_id = Counter(1)
        super(Sections, self).__init__(parent=parent)
        sections = [] if sections is None else sections
        self._units = units
        self._sections = sorted(sections, key=attrgetter("name"))

        def section_id_maker(section: Section) -> Section:
            if section.id is None:
                section.id = next(sec_id)
            return section

        [section_id_maker(sec) for sec in self._sections]

        self.recreate_name_and_id_maps(self._sections)

        if len(self._name_map.keys()) != len(self._id_map.keys()):
            import collections

            names = [sec.name for sec in self._sections]
            counts = collections.Counter(names)
            filtered_elements = {element: count for element, count in counts.items() if count > 1}
            logger.warning(f"The following sections are non-unique '{filtered_elements}'")

    def renumber_id(self, start_id=1):
        cnt = Counter(start=start_id)
        for mat_id in sorted(self.id_map.keys()):
            mat = self.get_by_id(mat_id)
            mat.id = next(cnt)
        self.recreate_name_and_id_maps(self._sections)

    def __contains__(self, item):
        return item.name in self._name_map.keys()

    def __len__(self):
        return len(self._sections)

    def __iter__(self) -> Iterable[Section]:
        return iter(self._sections)

    def __getitem__(self, index):
        result = self._sections[index]
        return Sections(result) if isinstance(index, slice) else result

    def __add__(self, other: Sections):
        if self.parent is None:
            logger.error(f'Parent is None for Sections container "{self}"')
        for sec in other:
            sec.parent = self.parent
        other.renumber_id(self.max_id + 1)
        return Sections(chain(self, other), parent=self.parent)

    def __repr__(self):
        rpr = reprlib.Repr()
        rpr.maxlist = 8
        rpr.maxlevel = 1
        return f"Sections({rpr.repr(self._sections) if self._sections else ''})"

    def merge_sections_by_properties(self):
        models = []
        final_sections = []
        for i, sec in enumerate(self.sections):
            if sec not in models:
                models.append(sec)
                final_sections.append(sec)
            else:
                index = models.index(sec)
                replacement_sec = models[index].parent
                for ref in sec.refs:
                    ref.section = replacement_sec

        self._sections = final_sections
        self.recreate_name_and_id_maps(self._sections)

    def index(self, item):
        index = bisect_left(self._sections, item)
        if (index != len(self._sections)) and (self._sections[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item: Section) -> int:
        return int(item in self)

    def get_by_name(self, name: str) -> Section:
        if name not in self._name_map.keys():
            raise ValueError(f'The section id "{name}" is not found')

        return self._name_map[name]

    def get_by_id(self, sec_id: int) -> Section:
        if sec_id not in self._id_map.keys():
            raise ValueError(f'The node id "{sec_id}" is not found')

        return self._id_map[sec_id]

    @property
    def id_map(self) -> dict[int, Section]:
        return self._id_map

    @property
    def name_map(self) -> dict[str, Section]:
        return self._name_map

    def add(self, section: Section) -> Section:
        if section.name is None:
            raise Exception("Name is not allowed to be None.")

        # Note: Evaluate if parent should be "Sections" not Part object?
        if section.parent is None:
            section.parent = self._parent

        if section in self._sections:
            index = self._sections.index(section)
            existing_section = self._sections[index]
            for elem in section.refs:
                elem.section = existing_section
                if elem not in existing_section.refs:
                    existing_section.refs.append(elem)
            return existing_section

        if section.name in self._name_map.keys():
            logger.info(f'Section with same name "{section.name}" already exists. Will use that section instead')
            existing_section = self._name_map[section.name]
            for elem in section.refs:
                elem.section = existing_section
                if elem not in existing_section.refs:
                    existing_section.refs.append(elem)
            return existing_section

        if section.id is None:
            section.id = self.max_id + 1

        if len(self._sections) > 0 and section.id in self._id_map.keys():
            section.id = self.max_id + 1

        self._sections.append(section)
        self._id_map[section.id] = section
        self._name_map[section.name] = section

        return section

    @property
    def sections(self) -> list[Section]:
        return self._sections

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            for m in self._sections:
                m.units = value
            self._units = value


class Nodes:
    def __init__(self, nodes=None, parent=None, from_np_array=None):
        self._parent = parent

        if from_np_array is not None:
            self._array = from_np_array
            nodes = self._np_array_to_nlist(from_np_array)
        else:
            nodes = [] if nodes is None else nodes

        self._nodes = list(nodes)

        if len(tuple(set(self._nodes))) != len(self._nodes):
            raise DuplicateNodes("Duplicate Nodes not allowed in a Nodes object")

        self._idmap = dict()
        self._bbox = None
        self._maxid = 0
        if len(self._nodes) > 0:
            self._sort()
            self._maxid = max(self._idmap.keys())
            self._bbox = self._get_bbox()

    def _sort(self):
        self._nodes = sorted(self._nodes, key=attrgetter("x", "y", "z"))
        try:
            self._idmap = {n.id: n for n in sorted(self._nodes, key=attrgetter("id"))}
        except TypeError as e:
            raise TypeError(e)

    def renumber(self, start_id: int = 1, renumber_map: dict = None):
        """Ensures that the node numberings starts at 1 and has no holes in its numbering."""
        if renumber_map is not None:
            self._renumber_from_map(renumber_map)
        else:
            self._renumber_linearly(start_id)

        self._sort()
        self._maxid = max(self._idmap.keys()) if len(self._nodes) > 0 else 0
        self._bbox = self._get_bbox() if len(self._nodes) > 0 else None

    def _renumber_linearly(self, start_id):
        for i, n in enumerate(sorted(self._nodes, key=attrgetter("id")), start=start_id):
            if i != n.id:
                n.id = i

    def _renumber_from_map(self, renumber_map):
        for n in sorted(self._nodes, key=attrgetter("id")):
            n.id = renumber_map[n.id]

    def _np_array_to_nlist(self, np_array):
        from ada import Node

        return [Node(row[1:], int(row[0]), parent=self._parent) for row in np_array]

    def to_np_array(self, include_id=False):
        if include_id:
            return np.array([(n.id, *n.p) for n in self._nodes])
        else:
            return np.array([n.p for n in self._nodes])

    def to_fem_nodes(self) -> FemNodes:
        from ada.fem.results.common import FemNodes

        node_refs = self.to_np_array(include_id=True)
        identifiers = node_refs[:, 0]
        coords = node_refs[:, 1:]

        return FemNodes(coords, identifiers)

    def __contains__(self, item):
        return item in self._nodes

    def __len__(self):
        return len(self._nodes)

    def __iter__(self) -> Iterable[Node]:
        return iter(self._nodes)

    def __getitem__(self, index):
        result = self._nodes[index]
        return Nodes(result) if isinstance(index, slice) else result

    def __eq__(self, other):
        if not isinstance(other, Nodes):
            return NotImplemented
        return self._nodes == other._nodes

    def __ne__(self, other):
        if not isinstance(other, Nodes):
            return NotImplemented
        return self._nodes != other._nodes

    def __add__(self, other: Nodes):
        for n in other.nodes:
            n.parent = self.parent
        return Nodes(chain(self._nodes, other.nodes))

    def __repr__(self):
        return f"Nodes({len(self._nodes)}, min_id: {self.min_nid}, max_id: {self.max_nid})"

    def index(self, item):
        index = bisect_left(self._nodes, item)
        if (index != len(self._nodes)) and (self._nodes[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def count(self, item):
        return int(item in self)

    def move(self, move: Iterable[float, float, float] = None, rotate: Rotation = None):
        """A method for translating and/or rotating your model."""

        def moving(no):
            no.p = no.p + move

        def map_rotations(no, p):
            no.p = p

        if rotate is not None:
            origin = np.array(rotate.origin)
            rot_mat = rotate.to_rot_matrix()
            vectors = np.array([n.p - origin for n in self._nodes])
            res = np.matmul(vectors, rot_mat.T)
            [map_rotations(n, p + origin) for n, p in zip(self._nodes, res)]

        if move is not None:
            move = np.array(move)
            list(map(moving, self._nodes))

        self._sort()

    def from_id(self, nid: int):
        if nid not in self._idmap.keys():
            raise ValueError(f'The node id "{nid}" is not found')
        else:
            return self._idmap[nid]

    def _get_bbox(self):
        if len(self._nodes) == 0:
            raise ValueError("No Nodes are found")
        nodes_yids = sorted(self._nodes, key=attrgetter("y"))
        nodes_zids = sorted(self._nodes, key=attrgetter("z"))
        xmin, xmax = self._nodes[0], self._nodes[-1]
        ymin, ymax = nodes_yids[0], nodes_yids[-1]
        zmin, zmax = nodes_zids[0], nodes_zids[-1]
        return (xmin, xmax), (ymin, ymax), (zmin, zmax)

    @property
    def dmap(self) -> Dict[str, Node]:
        return self._idmap

    @property
    def bbox(self):
        if self._bbox is None:
            self._bbox = self._get_bbox()
        return self._bbox

    @property
    def vol_cog(self):
        return tuple([(self.bbox[i][0][i] + self.bbox[i][1][i]) / 2 for i in range(3)])

    @property
    def max_nid(self) -> int:
        return max(self.dmap.keys()) if len(self.dmap.keys()) > 0 else 0

    @property
    def min_nid(self) -> int:
        return min(self.dmap.keys()) if len(self.dmap.keys()) > 0 else 0

    @property
    def nodes(self) -> List[Node]:
        return self._nodes

    def get_by_volume(self, p=None, vol_box=None, vol_cyl=None, tol=Settings.point_tol) -> List[Node]:
        """

        :param p: Point
        :param vol_box: Additional point to find nodes inside a rectangular box
        :param vol_cyl: (radius, height, cylinder thickness). Note! Radius is measured to outside of cylinder wall
        :param tol: Point tolerance
        :return:
        """
        p = np.array(p) if type(p) is (list, tuple) else p
        if p is not None and vol_cyl is None and vol_box is None:
            vol = [(coord - tol, coord + tol) for coord in p]
        elif vol_box is not None:
            vol = list(zip(p, vol_box))
        elif vol_cyl is not None and p is not None:
            r, h, t = vol_cyl
            vol = [
                (p[0] - r - tol, p[0] + r + tol),
                (p[1] - r - tol, p[1] + r + tol),
                (p[2] - tol, p[2] + tol + h),
            ]
        else:
            raise Exception("No valid search input provided. None is returned")

        vol_min, vol_max = zip(*vol)
        xmin = bisect_left(self._nodes, Node(vol_min))
        xmax = bisect_right(self._nodes, Node(vol_max))

        xlist = sorted(self._nodes[xmin:xmax], key=attrgetter("y"))
        ysorted = [n.y for n in xlist]
        ymin = bisect_left(ysorted, vol_min[1])
        ymax = bisect_right(ysorted, vol_max[1])

        ylist = sorted(xlist[ymin:ymax], key=attrgetter("z"))
        zsorted = [n.z for n in ylist]
        zmin = bisect_left(zsorted, vol_min[2])
        zmax = bisect_right(zsorted, vol_max[2])

        simplesearch = ylist[zmin:zmax]

        if vol_cyl is not None:
            r, h, t = vol_cyl
            pt1_ = p + np.array([0, 0, -h])
            pt2_ = p + np.array([0, 0, +h])

            def eval_p_in_cyl(no):
                if t == r:
                    if points_in_cylinder(pt1_, pt2_, r, no.p) is True:
                        return no
                else:
                    eval1 = points_in_cylinder(pt1_, pt2_, r + t, no.p)
                    eval2 = points_in_cylinder(pt1_, pt2_, r - t, no.p)
                    if eval1 is True and eval2 is False:
                        return no
                return None

            return list(filter(None, [eval_p_in_cyl(q) for q in simplesearch]))
        else:
            return list(simplesearch)

    def add(self, node: Node, point_tol: float = Settings.point_tol, allow_coincident: bool = False) -> Node:
        """Insert node into sorted list"""

        def insert_node(n, i):
            new_id = int(self._maxid + 1) if len(self._nodes) > 0 else 1
            if n.id in self._idmap.keys() or n.id is None:
                n.id = new_id

            self._nodes.insert(i, n)
            self._idmap[n.id] = n
            self._bbox = None
            self._maxid = n.id if n.id > self._maxid else self._maxid

        index = bisect_left(self._nodes, node)
        if (len(self._nodes) != 0) and allow_coincident is False:
            res = self.get_by_volume(node.p, tol=point_tol)
            if len(res) == 1:
                nearest_node = res[0]
                vlen = vector_length(nearest_node.p - node.p)
                if vlen < point_tol:
                    logger.debug(f'Replaced new node with node id "{nearest_node.id}" found within point tolerances')
                    return nearest_node

        insert_node(node, index)

        if node.parent is None:
            node.parent = self.parent

        return node

    def remove(self, nodes: Union[Node, Iterable[Node]]):
        """Remove node(s) from the nodes container"""
        nodes = list(nodes) if isinstance(nodes, Iterable) else [nodes]
        for node in nodes:
            if node in self._nodes:
                logger.debug(f"Removing {node}")
                self._nodes.pop(self._nodes.index(node))
                self.renumber()
            else:
                logger.error(f"'{node}' not found in node-container.")

    def remove_standalones(self) -> None:
        """Remove nodes that are without any usage references"""
        self.remove(filter(lambda x: not x.has_refs, self._nodes))

    def merge_coincident(self, tol: float = Settings.point_tol) -> None:
        """
        Merge nodes which are within the standard default of Nodes.get_by_volume. Nodes merged into the node connected
        to most elements.
        :return:
        """

        def replace_duplicate_nodes(duplicates: Iterable[Node], new_node: Node):
            if duplicates and len(new_node.refs) >= np.max(list(map(lambda x: len(x.refs), duplicates))):
                for duplicate_node in duplicates:
                    replace_node(duplicate_node, new_node)
                    self.remove(duplicate_node)

        for node in filter(lambda x: x.has_refs, self._nodes):
            duplicate_nodes = list(
                sorted(
                    filter(lambda x: x.id != node.id, self.get_by_volume(node.p, tol=tol)), key=lambda x: len(x.refs)
                )
            )
            replace_duplicate_nodes(duplicate_nodes, node)

        self._sort()

    def rounding_node_points(self, precision: int = Settings.precision) -> None:
        """Rounds all nodes to set precision"""
        for node in self.nodes:
            node.p_roundoff(precision=precision)

    @property
    def parent(self) -> Union[Part, FEM]:
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, List, Union

import numpy as np
from pyquaternion import Quaternion

if TYPE_CHECKING:
    from ada import Part
    from ada.base.physical_objects import BackendGeom


@dataclass
class Transform:
    translation: np.ndarray = None
    rotation: Rotation = None

    def to_ifc(self, f):
        from ada.ifc.utils import export_transform

        return export_transform(f, self)


@dataclass
class Rotation:
    origin: Iterable[float, float, float]
    vector: Iterable[float, float, float]
    angle: float

    def to_rot_matrix(self):
        my_quaternion = Quaternion(axis=self.vector, degrees=self.angle)
        return my_quaternion.rotation_matrix

    def rotate_point(self, p: Union[tuple, list]):
        p1 = np.array(self.origin)
        rot_mat = self.to_rot_matrix()
        p_norm = np.array(p) - p1
        res = p1 + p_norm @ rot_mat.T
        return res


@dataclass
class Placement:
    origin: Union[list, tuple, np.ndarray] = (0, 0, 0)
    xdir: Union[list, tuple, np.ndarray] = None
    ydir: Union[list, tuple, np.ndarray] = None
    zdir: Union[list, tuple, np.ndarray] = None
    scale: float = 1.0
    parent = None

    def __post_init__(self):
        from ada.core.vector_utils import calc_yvec

        all_dir = [self.xdir, self.ydir, self.zdir]
        if all(x is None for x in all_dir):
            self.xdir = np.array([1, 0, 0], dtype=float)
            self.ydir = np.array([0, 1, 0], dtype=float)
            self.zdir = np.array([0, 0, 1], dtype=float)

        if self.ydir is None and all(x is not None for x in [self.xdir, self.zdir]):
            self.ydir = calc_yvec(self.xdir, self.zdir)

        all_dir = [self.xdir, self.ydir, self.zdir]

        if all(x is None for x in all_dir):
            raise ValueError("Placement orientation needs all 3 vectors")

        self.xdir = np.array(self.xdir, dtype=float)
        self.ydir = np.array(self.ydir, dtype=float)
        self.zdir = np.array(self.zdir, dtype=float)

        if not isinstance(self.origin, np.ndarray):
            self.origin = np.array(self.origin, dtype=float)

    def absolute_placement(self):
        current_location = np.array([0, 0, 0], dtype=float)
        ancestry = self.parent.get_ancestors()
        ancestry.reverse()
        for ancestor in ancestry:
            current_location += ancestor.placement.origin
            # TODO: Add support for combining rotations as well
        return current_location

    def to_vector_geom(self, **kwargs) -> "Part":
        from ada.occ.utils import make_ori_vector

        return make_ori_vector("VecGeom", self.origin, self.csys, **kwargs)

    @property
    def csys(self):
        return [self.xdir, self.ydir, self.zdir]

    def __eq__(self, other: Placement):
        from ada.core.vector_utils import vector_length

        for prop in ["origin", "xdir", "ydir", "zdir"]:
            if vector_length(getattr(other, prop) - getattr(self, prop)) > 0.0:
                return False

        return True


@dataclass
class Instance:
    instance_ref: Union["Part", "BackendGeom"]
    placements: List[Placement] = field(default_factory=list)

    def to_list_of_custom_json_matrices(self):
        from pyquaternion import Quaternion

        from ada.ifc.utils import create_guid

        matrices = [[self.instance_ref.guid, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]]
        for place in self.placements:
            q1 = Quaternion(matrix=np.array(place.csys))
            rmat = q1.rotation_matrix
            matrices.append(
                [create_guid(), *place.origin.astype(float).tolist(), *np.concatenate(rmat).astype(float).tolist()]
            )

        return matrices



from ada import Wall
from ada.base.units import Units
from ada.core.constants import O, X, Z
from ada.ifc.utils import (
    create_guid,
    create_ifc_placement,
    create_ifcextrudedareasolid,
    create_ifcpolyline,
    create_local_placement,
    tesselate_shape,
)

from .write_stru_components import write_door, write_window


def write_ifc_wall(wall: Wall):
    if wall.parent is None:
        raise ValueError("Ifc element cannot be built without any parent element")

    a = wall.parent.get_assembly()
    ifc_store = a.ifc_store
    f = ifc_store.f

    owner_history = ifc_store.owner_history
    parent = f.by_guid(wall.parent.guid)
    elevation = wall.placement.origin[2]

    # Wall creation: Define the wall shape as a polyline axis and an extruded area solid
    wall_placement = create_local_placement(f, relative_to=parent.ObjectPlacement)

    polyline2d = create_ifcpolyline(f, wall.points)
    axis_representation = f.createIfcShapeRepresentation(ifc_store.get_context("Axis"), "Axis", "Curve2D", [polyline2d])

    extrusion_placement = create_ifc_placement(f, (0.0, 0.0, float(elevation)), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0))

    polyline = create_ifcpolyline(f, wall.extrusion_area)
    profile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)

    solid = create_ifcextrudedareasolid(f, profile, extrusion_placement, (0.0, 0.0, 1.0), wall.height)
    body = f.createIfcShapeRepresentation(ifc_store.get_context("Body"), "Body", "SweptSolid", [solid])

    product_shape = f.createIfcProductDefinitionShape(None, None, [axis_representation, body])

    wall_el = f.create_entity(
        "IfcWall",
        GlobalId=wall.guid,
        OwnerHistory=owner_history,
        Name=wall.name,
        Description="An awesome wall",
        ObjectType=None,
        ObjectPlacement=wall_placement,
        Representation=product_shape,
    )

    return wall_el


def add_ifc_insert_elem(wall: Wall, shape_, opening_element, wall_el, ifc_type):
    a = wall.parent.get_assembly()
    ifc_store = a.ifc_store
    f = ifc_store.f

    owner_history = ifc_store.owner_history
    schema = f.wrapped_data.schema

    # Create a simplified representation for the Window
    insert_placement = create_local_placement(f, O, Z, X, wall_el.ObjectPlacement)

    shape = shape_.geom

    insert_shape_ = tesselate_shape(shape, schema, Units.get_general_point_tol(a.units))
    insert_shape = f.add(insert_shape_)

    # Link to representation context
    body_context = ifc_store.get_context("Body")
    for rep in insert_shape.Representations:
        rep.ContextOfItems = body_context

    insert_map = dict(IfcWindow=write_window, IfcDoor=write_door)

    insert_writer = insert_map.get(ifc_type, None)

    if insert_writer is None:
        raise ValueError(f'Currently unsupported ifc_type "{ifc_type}"')

    ifc_insert = insert_writer(f, owner_history, insert_placement, insert_shape)

    # Relate the window to the opening element
    f.create_entity(
        "IfcRelFillsElement",
        create_guid(),
        owner_history,
        None,
        None,
        opening_element,
        ifc_insert,
    )
    return ifc_insert



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Callable

from ada.base.root import Root
from ada.base.types import GeomRepr
from ada.base.units import Units
from ada.concepts.transforms import Placement
from ada.core.constants import color_map as _cmap
from ada.occ.store import OCCStore
from ada.visualize.config import ExportConfig

if TYPE_CHECKING:
    from ada import FEM, Penetration
    from ada.fem import Elem
    from ada.fem.meshing import GmshOptions
    from ada.ifc.store import IfcStore


# TODO: Consider storing primitive geometry definitions as an attribute in the BackendGeom class to simplify subclasses.
class BackendGeom(Root):
    """The backend of all physical components (Beam, Plate, etc.) or aggregate of components (Part, Assembly)"""

    _renderer = None

    def __init__(
        self,
        name,
        guid=None,
        metadata=None,
        units=Units.M,
        parent=None,
        colour=None,
        placement=Placement(),
        ifc_store: IfcStore = None,
        opacity=1.0,
    ):
        super().__init__(name, guid, metadata, units, parent, ifc_store=ifc_store)
        from ada.visualize.new_render_api import Visualize

        self._penetrations = []
        self._placement = placement
        placement.parent = self
        self.colour = colour
        self.opacity = opacity
        self._elem_refs = []
        self._viz = Visualize(self)

    def add_penetration(self, pen, add_to_layer: str = None):
        from ada import Penetration, Shape
        from ada.base.changes import ChangeAction

        pen.parent = self

        if issubclass(type(pen), Shape) is True:
            pen = Penetration(pen, parent=self)
            self._penetrations.append(pen)
        elif type(pen) is Penetration:
            self._penetrations.append(pen)
        else:
            raise ValueError("")

        if self.change_type in (ChangeAction.NOCHANGE, ChangeAction.NOTDEFINED):
            self.change_type = ChangeAction.MODIFIED

        if add_to_layer is not None:
            a = self.get_assembly()
            a.presentation_layers.add_object(pen, add_to_layer)

        return pen

    def to_fem_obj(
        self,
        mesh_size,
        geom_repr: str | GeomRepr,
        options: GmshOptions = None,
        silent=True,
        use_quads=False,
        use_hex=False,
        name="AdaFEM",
        interactive=False,
    ) -> FEM:
        from ada.fem.meshing import GmshOptions, GmshSession

        if isinstance(geom_repr, str):
            geom_repr = GeomRepr.from_str(geom_repr)

        options = GmshOptions(Mesh_Algorithm=8) if options is None else options
        with GmshSession(silent=silent, options=options) as gs:
            gs.add_obj(self, geom_repr=geom_repr)
            gs.mesh(mesh_size, use_quads=use_quads, use_hex=use_hex)
            if interactive:
                gs.open_gui()
            return gs.get_fem(name)

    def to_fem(
        self,
        mesh_size,
        geom_repr,
        name: str,
        fem_format: str,
        options: GmshOptions = None,
        silent=True,
        use_quads=False,
        use_hex=False,
        return_assembly=False,
        **kwargs,
    ):
        from ada import Assembly, Part

        p = Part(name)
        p.fem = self.to_fem_obj(mesh_size, geom_repr, options, silent, use_quads, use_hex, name)
        a = Assembly() / (p / self)
        if return_assembly:
            return a
        a.to_fem(name, fem_format, **kwargs)

    def to_stp(self, destination_file, geom_repr: GeomRepr = GeomRepr.SOLID, progress_callback: Callable = None):
        step_writer = OCCStore.get_writer()
        step_writer.add_shape(self.solid(), self.name, rgb_color=self.colour_norm)
        step_writer.export(destination_file)

    def to_obj_mesh(self, geom_repr: str | GeomRepr = GeomRepr.SOLID, export_config: ExportConfig = ExportConfig()):
        from ada.visualize.formats.assembly_mesh.write_objects_to_mesh import (
            occ_geom_to_poly_mesh,
        )

        if isinstance(geom_repr, str):
            geom_repr = GeomRepr.from_str(geom_repr)
        return occ_geom_to_poly_mesh(self, geom_repr=geom_repr, export_config=export_config)

    def render_locally(
        self, addr="localhost", server_port=8080, open_webbrowser=False, render_engine="threejs", resolution=(1800, 900)
    ):
        from OCC.Display.WebGl.simple_server import start_server

        from ada.visualize.renderer_pythreejs import MyRenderer

        if render_engine == "xdom":
            from OCC.Display.WebGl import x3dom_renderer

            my_renderer = x3dom_renderer.X3DomRenderer()
            # TODO: Find similarities in build processing as done for THREE.js (tesselate geom etc..).
            # my_renderer.DisplayShape(shape.profile_curve_outer.wire)
            # my_renderer.DisplayShape(shape.sweep_curve.wire)
            # my_renderer.DisplayShape(shape.geom)
            my_renderer.render()
        else:  # Assume THREEJS
            from ipywidgets.embed import embed_minimal_html

            _path = pathlib.Path("temp/index.html").resolve().absolute()

            renderer = MyRenderer(resolution)
            renderer.DisplayObj(self)
            renderer.build_display()

            os.makedirs(_path.parent, exist_ok=True)
            embed_minimal_html(_path, views=renderer.renderer, title="Pythreejs Viewer")
            start_server(addr, server_port, str(_path.parent), open_webbrowser)

    def get_render_snippet(self, view_size=None):
        """Return the html snippet containing threejs renderer"""
        from ipywidgets.embed import embed_snippet

        from ada.visualize.renderer_pythreejs import MyRenderer

        renderer = MyRenderer()
        renderer.DisplayObj(self)
        renderer.build_display()

        return embed_snippet(renderer.renderer)

    @property
    def colour(self):
        return self._colour

    @colour.setter
    def colour(self, value):
        if type(value) is str:
            if value.lower() not in _cmap.keys():
                raise ValueError("Currently unsupported")
            self._colour = _cmap[value.lower()]
        else:
            self._colour = value

    @property
    def colour_norm(self):
        if self._colour is None:
            self.colour = "white"
        return [x / 255 for x in self.colour] if any(i > 1 for i in self.colour) else self.colour

    @property
    def colour_webgl(self):
        from OCC.Display.WebGl.jupyter_renderer import format_color

        if self.colour is None:
            return None
        if self.colour[0] == -1 and self.colour[1] == -1 and self.colour[2] == -1:
            return None

        if self.colour[0] <= 1.0:
            colour = [int(x * 255) for x in self.colour]
        else:
            colour = [int(x) for x in self.colour]

        colour_formatted = format_color(*colour)
        return colour_formatted

    @property
    def opacity(self):
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        if (0.0 <= value <= 1.0) is False:
            raise ValueError(f'Opacity is only valid between 1 and 0. "{value}" was passed in')

        self._opacity = value

    @property
    def transparent(self):
        return False if self.opacity == 1.0 else True

    @property
    def penetrations(self) -> list[Penetration]:
        return self._penetrations

    @property
    def elem_refs(self) -> list[Elem]:
        return self._elem_refs

    @elem_refs.setter
    def elem_refs(self, value):
        self._elem_refs = value

    @property
    def placement(self) -> Placement:
        return self._placement

    @placement.setter
    def placement(self, value: Placement):
        self._placement = value

    def _repr_html_(self):
        from ada.config import Settings

        if Settings.use_new_visualize_api is True:
            self._viz.objects = []
            self._viz.add_obj(self)
            self._viz.display(return_viewer=False)
            return ""

        from IPython.display import display
        from ipywidgets import HBox, VBox

        from ada.visualize.renderer_pythreejs import MyRenderer

        renderer = MyRenderer()

        renderer.DisplayObj(self)
        renderer.build_display()
        self._renderer = renderer
        display(HBox([VBox([HBox(renderer.controls), renderer.renderer]), renderer.html]))
        return ""

    def solid(self):
        raise NotImplementedError()

    def shell(self):
        raise NotImplementedError()

    def line(self):
        raise NotImplementedError()



import numpy as np

from ada.core.utils import roundoff


def import_indexedpolycurve(ipoly, normal, xdir, origin):
    """

    :param ipoly: IFC element
    :param normal:
    :param xdir:
    :param origin:
    :return:
    """
    from ada import ArcSegment, LineSegment
    from ada.core.curve_utils import segments_to_local_points
    from ada.core.vector_utils import global_2_local_nodes

    ydir = np.cross(normal, xdir)
    nodes3d = [p for p in ipoly.Points.CoordList]
    nodes2d = global_2_local_nodes([xdir, ydir], origin, nodes3d)
    nodes2d = [np.array([n[0], n[1], 0.0]) for n in nodes2d]
    seg_list = []
    for i, seg in enumerate(ipoly.Segments):
        if seg.is_a("IfcLineIndex"):
            v = seg.wrappedValue
            p1 = nodes2d[v[0] - 1]
            p2 = nodes2d[v[1] - 1]
            seg_list.append(LineSegment(p1=p1, p2=p2))
        elif seg.is_a("IfcArcIndex"):
            v = seg.wrappedValue
            p1 = nodes2d[v[0] - 1]
            p2 = nodes2d[v[1] - 1]
            p3 = nodes2d[v[2] - 1]
            seg_list.append(ArcSegment(p1, p3, midpoint=p2))
        else:
            raise ValueError("Unrecognized type")

    local_points = [(roundoff(x[0]), roundoff(x[1])) for x in segments_to_local_points(seg_list)]
    return local_points


def import_polycurve(poly, normal, xdir):
    from ada.core.vector_utils import global_2_local_nodes

    ydir = np.cross(normal, xdir)
    nodes3d = [p for p in poly.Points]
    nodes2d = global_2_local_nodes([xdir, ydir], (0, 0, 0), nodes3d)

    return nodes2d



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ada.ifc.utils import create_guid
from ada.visualize.concept import ObjectMesh, PartMesh, VisMesh

if TYPE_CHECKING:
    from ada.fem.results import Results


def export_results_to_assembly_mesh(results: "Results", data_type) -> VisMesh:
    name = results.assembly.name

    res_mesh = results.result_mesh
    data = np.asarray(res_mesh.mesh.point_data[data_type], dtype="float32")
    vertices = np.asarray([x + u[:3] for x, u in zip(res_mesh.vertices, data)], dtype="float32")
    colors = res_mesh.colorize_data(data)
    faces = res_mesh.faces
    guid = create_guid(name)
    id_map = {
        guid: ObjectMesh(
            guid=guid,
            faces=faces.astype(int),
            position=vertices.flatten().astype(float),
            normal=None,
            color=None,
            vertex_color=colors.flatten().astype(float).tolist(),
            instances=None,
        )
    }
    pm = PartMesh(name=name, id_map=id_map)
    project = results.assembly.metadata.get("project", "DummyProject")
    return VisMesh(name=name, project=project, world=[pm], meta=None)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import Part


@dataclass
class SATRefs:
    sat_text: str
    sat_map: dict


def create_sat_from_beams(part: Part) -> SATRefs:
    from ada import Beam

    sat_str = ""
    sat_map = dict()
    for bm in part.get_all_physical_objects(by_type=Beam):
        pass

    return SATRefs(sat_str, sat_map)
